/* 
  File: calc.c

   Annealing schedules:
   
     EXP:       T(k+1) = alpha * T(K)

     ADAPTIVE:  T(k+1) = T(k) * exp [ -lambda * T(k) / sd ( E ) ],

                where sd ( E ) is the standard deviation of the ensemble
		of energies of size Ensemble.

 */

#include "calc.h"


int    gl_count = 0,
       Num_Reject[MAX_BETAS],
       Num_Accept[MAX_BETAS],
       Measurements;

double Total_Energy, Total_Magnet, Total_Accept,
       Tot_Energy_Sq, Tot_Magnet_Sq, Tot_Accept_Sq,  
       Error_Energy, Error_Magnet, Error_Accept, 
       Std_Dev = INFIN,
       Min_Energy[MAX_BETAS];

extern FILE   
       *Efile;

extern int
       Cur_Beta_Index,
       NTherm, Num_Sweeps, Freq,
       Spins[][],       /* spins */
       J[][][];     /* couplings */

extern double
       Init_Beta (),
       t_rand ();

static double 
       find_avg (),
       find_var ();

static void
       calculate_cost (),
       calculate_sd (),
       metropolis_update (),
       thermalize ();
   

double Get_New_Beta ( old_beta )
/*
  Adaptive annealing schedule from:

    G. Mirkin, K. Vasudevan, W. G. Laidlaw W. G. Wilson and F. A. Cook, 
    A comparison of several cooling schedules for simulated annealing 
    implemented on a residual statistics problem, Geophyis Res Lett, vol
    20, no 1, 77-80, Jan 8, 1993.

  Here beta is 1 / Temp.
  The adaptive beta selection works like this:

        T(k+1) = T(k) * exp [ -lambda * T(k) / sd ( E ) ]

  or in terms of beta,

	B(k+1) = B(k) / exp [ -lambda / B(k) / sd ( E ) ].

  Default annealing schedule is exponential.

 */
double old_beta;
{

  switch ( SCHED )
  {
    case ADAPT :
      return ( old_beta / exp( - Lambda / old_beta / Std_Dev ) );

    default :
      return old_beta / Alpha;
  }

}


int Get_Iter( n )
/* 
   Input: Number of beta subdivisions.
  Output: Number of iterations.  

    Info: Here, 'iteration' refers to one pass over the entire lattice.  
          Various single passes are negligible.  E.g. calculate_cost(), 
	  Save_Spins(), Update_Config().
 */
int n;
{

  if ( SCHED == ADAPT )
    return ( Cur_Beta_Index * n * (NTherm + (Num_Sweeps * Freq) + Ensemble) );

  else 
    return ( Cur_Beta_Index * n * (NTherm + (Num_Sweeps * Freq)) );

}


void Calc_Cost ( beta, energy, magnet, accept, FLAG )
/* 
   Input:
  Output:

    Misc:
 */
double beta, *energy, *accept, *magnet;
int    FLAG;
{
  int lattice;

  for ( lattice = 0; lattice < Freq; lattice ++ ) 
    metropolis_update( beta, accept, FLAG );

  calculate_cost( energy, magnet );

}


void One_Pass ( beta, min_beta, overlap_coeff )
/* 
   Input:
  Output:

    Misc:
 */

double beta, *min_beta, overlap_coeff;
{
  int    sweep, flag, n_sweeps,
         temp_spins[MAXROWS][MAXCOLS];

  double energy,
         accept,
         magnet;


  Total_Energy = 0.0;
  Total_Magnet = 0.0;
  Total_Accept = 0.0;

  Tot_Energy_Sq = 0.0;
  Tot_Magnet_Sq = 0.0;
  Tot_Accept_Sq = 0.0;

  flag = ON;
  thermalize( beta, &accept, flag );

/*  n_sweeps = Num_Sweeps + (Num_Sweeps * fabs( overlap_coeff ));

*/
  n_sweeps = Num_Sweeps;
  Measurements = n_sweeps;

  for ( sweep = 0; sweep < n_sweeps; sweep++ ) {
    Calc_Cost( beta, &energy, &magnet, &accept, flag );

    if ( energy < Min_Energy[Cur_Beta_Index] ) {
      Min_Energy[Cur_Beta_Index] = energy;
      *min_beta = beta;
    }

    Total_Energy += energy;
    Total_Magnet += magnet;
    Total_Accept += accept;
 
    Tot_Energy_Sq += energy * energy;
    Tot_Magnet_Sq += magnet * magnet;
    Tot_Accept_Sq += accept * accept;

    if ( DUMP_ENERGY )
      if ( ! (sweep % 5) ) {
	fprintf( Efile, "%d %lf\n", gl_count ++, energy );
	fflush ( Efile );	 
      }
  }

/* The system (should be) decorralated at this point and we 
   are ready to move to a new beta.  If the adaptive sched is
   used, calculate the standard deviation.
 */
  if ( SCHED == ADAPT ) 
    calculate_sd ( beta );
   
  Total_Energy /= Measurements;
  Total_Magnet /= Measurements;
  Total_Accept /= Measurements;

  Tot_Energy_Sq /= Measurements;
  Tot_Magnet_Sq /= Measurements;
  Tot_Accept_Sq /= Measurements;

  Error_Energy = Tot_Energy_Sq - (Total_Energy * Total_Energy);
  Error_Magnet = Tot_Magnet_Sq - (Total_Magnet * Total_Magnet);
  Error_Accept = Tot_Accept_Sq - (Total_Accept * Total_Accept);

  Error_Energy /= Num_Spins;
  Error_Magnet /= Num_Spins;
  Error_Accept /= Num_Spins;

}   /* One_Pass ()  */


void thermalize( beta, accept, flag )
int    flag;
double beta,
       *accept;
{
  int sweep;
  
  for ( sweep = 0; sweep < NTherm; sweep++ )
    metropolis_update (  beta, accept, flag );

}


void calculate_sd( beta )
/* 
   Input: Current beta
  Output: Standard Deviation calculated according to the following
          scheme.

  Ensemble computation for SD using n identical and independent copies 
  of the current system configuration which share only the temperature.

  See: Salamon, P. et al, Simulated Annealing with constant thermodynamic
       speed, Computer Physics Communications, vol 49, 423-428, 1988.

    Misc: When SD approaches zero, it is manually increased so that 
          subsequent betas don't grow abruptly.
 */
double beta;
{
  double sample_avg,
         sample_var,
         prev_SD,
         accept, energy, magnet,
         Energy[MAX_BETAS];

  int    i, 
         lattice,
         flag,

         cur_Spins[MAXROWS][MAXCOLS],
         cur_J[MAXROWS][MAXCOLS][2];

  flag  = OFF;
  prev_SD = Std_Dev;

  Save_Spins( cur_Spins );
  Save_Couplings( cur_J );

  for ( i = 0; i < Ensemble; i ++ ) {
    Calc_Cost( beta, &energy, &magnet, &accept, flag );
    Energy[i] = energy;
    Update_Config( cur_Spins, cur_J );
  }

  sample_avg = find_avg( Energy, Ensemble );
  sample_var = find_var( Energy, sample_avg, Ensemble );
  Std_Dev    = sqrt( sample_var );

  if ( Std_Dev < MIN_SD ) 
    return;
/*
{
    Std_Dev = prev_SD + (prev_SD * 0.5);
    DUMP_SD = OFF;
    if ( CONV )
      printf( "***" );
  }
*/
  if ( CONV )
    printf( "SD = %lf\n", Std_Dev );

} /* calculate_sd () */


void Save_Spins( cur_Spins )
/* 
  Saves current system configuration in Spins[][] into
  cur_Spins[][].
 */
int      cur_Spins[MAXROWS][MAXCOLS];
{
  int i, j;

  for ( i = 0; i < Rows; i ++ )
    for ( j = 0; j < Cols; j ++ ) 
      cur_Spins[i][j] = Spins[i][j];

} /* Save_Spins () */


void Save_Couplings( cur_J )
int      cur_J[MAXROWS][MAXCOLS][NDIM];
{
  int i, j, dim;

  for ( i = 0; i < Rows; i ++ )
    for ( j = 0; j < Cols; j ++ ) 
      for ( dim = 0; dim < NDIM; dim ++ )
	cur_J[i][j][dim] = J[i][j][dim];

} /* Save_Couplings () */


void Update_Config( cur_Spins, cur_J )
/*
  Updates system configuration Spins[][] and J[][][] to the configuration
  stored in cur_Spins[][] and cur_J[][][] respectively.
 */
int      cur_Spins[MAXROWS][MAXCOLS],
         cur_J[MAXROWS][MAXCOLS][2];
{
  int i, j, dim;
  
  for ( i = 0; i < Rows; i ++ )
    for ( j = 0; j < Cols; j ++ ) {
      Spins[i][j] = cur_Spins[i][j];
      for ( dim = 0; dim < 2; dim ++ )
	J[i][j][dim] = cur_J[i][j][dim];
    }

} /* Update_Config () */


double find_avg( Energy, size )
/* 
   Input: Energy array and sample size.
  Output: Average of energy values in Energy array.
 */
double Energy[MAX_BETAS];
int    size;
{
  double sum;
  int    i;

  sum   = 0.0;
  for ( i = 0; i < size; i ++ ) 
    sum += Energy[i];

  return sum / size;

} /* find_avg () */


double find_var( Energy, avg, size )
/*
   Input: Energy array, average of the energies, and sample size.
  Output: Variance of the sample.
 */
double avg,
       Energy[MAX_BETAS];
int    size;
{
  double var,
         dev;
  int    i;

  var = 0.0;
  for ( i = 0; i < size; i ++ )
  {
    dev = avg - Energy[i];
    var += dev * dev;
  }

  return var;
} /* find_var () */


double Init_Beta ( beta, config, model )
/* 
   Input:
  Output:

    Misc:
 */
double beta;
int    config, model;
{
#define    AUTO_START_EPS     10.0
#define    MAX_AUTO_ITER      10

  char   response;
  int    lattice, i,
         flag;
  double accept, energy, magnet, 
         percent_acc, factor, temp_beta;


  flag   = OFF;
  factor = 0.1;
  temp_beta = beta;
  for ( i = 0; i < MAX_AUTO_ITER; i ++ ) {
    for ( lattice = 0; lattice < Freq; lattice ++ )
      metropolis_update( temp_beta, &accept, flag );
    Init_Lattice ( config, model );

    percent_acc = accept / Num_Spins * 100.0;
    if ( fabs( percent_acc - 50.0 ) < AUTO_START_EPS ) {
      printf( "Dynamic Beta initialization info -> \n" );
      printf( "\t\tBeta = %lf\n", temp_beta );
      printf( "\t\tIterations = %d\n", i );

      return temp_beta;
    }
    temp_beta += temp_beta * factor;
  } 

  printf( "\nDynamic Beta determination failed...\n" );
  printf( "Current Beta is %lf...  Continue (n to EXIT) ? -> ", temp_beta );
  scanf ( "%c", &response );
  if ( (response == 'n') || (response == 'N') ) 
    exit (0);

} /* Init_Beta () */


int Terminate( step )
/* 
   Input:
  Output:

    Misc:
 */
int step;
{ 

  int    reject_diff,
         accept_diff;
  double min_diff = INFIN;

  /* visit at least step number of betas */
  if ( Cur_Beta_Index < step )
    return 0;

  reject_diff = abs ( Num_Reject[Cur_Beta_Index - 1] - 
		      Num_Reject[Cur_Beta_Index] );
  accept_diff = abs ( Num_Accept[Cur_Beta_Index - 1] - 
		      Num_Accept[Cur_Beta_Index] );
  min_diff    = fabs( Min_Energy[Cur_Beta_Index - step] - 
		      Min_Energy[Cur_Beta_Index] );

  if ( CONV )
    printf( "min diff = %lf \trej diff = %d \taccept diff = %d\n", 
	    min_diff, reject_diff, accept_diff );

  if ( (min_diff < MIN_EPS) || (accept_diff < ACCEPT_EPS) ) {
    if ( accept_diff < ACCEPT_EPS )
      printf( "Acceptance criteria satisfied...\n" );
    else
      printf( "Minimum Energy criteria satisfied...\n" );

    if ( (min_diff < MIN_EPS) && (accept_diff < ACCEPT_EPS) )
      return 1;
  }
  if ( Std_Dev < MIN_SD ) {
    printf( "Standard Deviation criteria satisfied...\n" );
    return 1;
  }
  return 0;

}  /* Terminate () */



/* PROCEDURE CALCULATE */
void  calculate_cost ( Energy, Magnet )
double *Energy, *Magnet;

{
   int Sum_Spins;
   int i, j;
   int ipos, jpos;

   *Magnet = 0.0;
   *Energy = 0.0;

   for ( i = 0; i < Rows; i++ )
   {
      if (i < Rows-1) 
         ipos = i+1;
      else 
         ipos = 0;
 
      for ( j = 0; j < Cols; j++ ) 
      {
         if (j < Cols-1)
            jpos = j+1;
         else 
            jpos = 0;
               
         Sum_Spins = Spins[i][j] * ( J[i][j][0] * Spins[ipos][j] + 
				     J[i][j][1] * Spins[i][jpos] );

         *Magnet += (double) Spins[i][j];
         *Energy += -1.0 * (double) Sum_Spins;
      }
   }

}

/* PROCEDURE METROPOLIS_UPDATE */
void  metropolis_update ( beta, Accept, flag )
double beta,
       *Accept;
int    flag;
{
   int Old_Spin, New_Spin, Spin_Sum;
   double Old_Energy, New_Energy, Energy_Diff;
   int Acceptance;
   int i, j;
   int ipos, jpos, ineg, jneg;

   *Accept = 0.0;
   
   for ( i = 0; i < Rows; i++ )
   {
      if (i < Rows-1) 
         ipos = i+1;
      else 
         ipos = 0;
      if (i > 0)
         ineg = i-1;
      else 
         ineg = Rows-1;
      for ( j = 0; j < Cols; j++ ) 
      {
         if (j < Cols-1)
            jpos = j+1;
         else 
            jpos = 0;
         if (j > 0) 
            jneg = j-1;
         else 
            jneg = Cols-1;

         Acceptance = 0;

         Old_Spin = Spins[i][j];
         New_Spin = -Old_Spin;

         Spin_Sum = J[i][j][0]    * Spins[ipos][j] + 
		    J[i][j][1]    * Spins[i][jpos] + 
		    J[ineg][j][0] * Spins[ineg][j] + 
		    J[i][jneg][1] * Spins[i][jneg] ;

	 /* These calculations are local, i.e. immediate neighbors of
	    the current lattice site, (i,j).
	  */
         Old_Energy = -beta * Old_Spin * Spin_Sum;
         New_Energy = -beta * New_Spin * Spin_Sum;
         Energy_Diff = New_Energy - Old_Energy;

         if ( (Energy_Diff <= 0.0) || ( exp(-Energy_Diff) > t_rand() ) )
         { 
            Spins[i][j] = New_Spin;
            Acceptance = Acceptance + 1;
	    if ( (New_Energy > Old_Energy) && flag ) 
	      Num_Accept[Cur_Beta_Index] ++;
         }
	 else if ( (New_Energy > Old_Energy) && flag ) 
	   Num_Reject[Cur_Beta_Index] ++;
	    
         *Accept += (double)Acceptance;
      }
   }
} /* metropolis_update () */



double Calc_Overlap( config1, config2 )
/* 
   Input: Two configurations.
  Output: A quantitative statement about the overlap of two configurations.

          -1.0 -> Corresponding configuration sites have opposite values.
	   1.0 -> Corresponding configuration sites have same values.

    Misc: Computes the overlap of two configurations.  Rows and Cols
          must be global variables.
 */
int    config1[MAXCOLS][MAXROWS],
       config2[MAXCOLS][MAXROWS];
{

  int    i, j,
         config_sum,
         vol;

  vol        = Rows * Cols;
  config_sum = 0;

  for ( i = 0; i < Rows; i ++ )
    for ( j = 0; j < Cols; j ++ )
      config_sum += config1[i][j] * config2[i][j];

  return (double)config_sum / (double)vol;

}


double drand48();

double t_rand()
{
 double temp;

/* temp = (double)rand()/(double) RAND_MAX; */
 temp = drand48();
 if (temp == 1.0) temp = 0.0;

 if ( (temp >1.0) || (temp < 0 ) ) {
     perror("ERROR: t_rand is bad");
     printf("trand = %lf \n",temp);
     exit(-2);
 }

 return(temp);

}
