! A.2 Hough Transformation

! The Hough transformation takes an input binary image of edge
! pixels and produces an array of edge parameters, indexed by its
! angle and radial distance from the origin. 

      PROGRAM HoughProg
      USE HPF_LIBRARY
      IMPLICIT NONE
      INTEGER, PARAMETER :: NN = 512
      INTEGER, PARAMETER :: STDOUT = 6
      INTEGER, PARAMETER :: FILEIN = 20, FILEOUT=21
      REAL, PARAMETER :: PId2 = 1.5707963
C .. Define interface for subroutines ..
      INTERFACE
        SUBROUTINE timer(return_time, initial_time)
          REAL, INTENT(IN)  :: initial_time
          REAL, INTENT(OUT) :: return_time
        END SUBROUTINE timer
      END INTERFACE
C .. Arrays and scalars ..
      INTEGER, DIMENSION(NN,NN) :: Hough
      INTEGER, DIMENSION(:,:), ALLOCATABLE :: Edge
      INTEGER, DIMENSION(:,:), ALLOCATABLE :: XX,YY
      INTEGER, DIMENSION(:,:), ALLOCATABLE :: iRho
      INTEGER, DIMENSION(:,:), ALLOCATABLE :: iRhoInd2
      INTEGER, DIMENSION(:), ALLOCATABLE :: XXpacked, YYpacked
      INTEGER, DIMENSION(:,:), ALLOCATABLE :: CountBase
      LOGICAL, DIMENSION(:,:), ALLOCATABLE :: CountMask
      INTEGER :: i,j,k
      INTEGER :: EsizeX, EsizeY
      INTEGER :: NedgePixels
      INTEGER :: iargc
      REAL, DIMENSION(NN) :: COStheta, SINtheta
      REAL :: dTheta, dRho
      REAL :: ti,tf
      REAL :: tstart,tend
      REAL :: tcomms, tcomp
      CHARACTER*(32) magic
      CHARACTER*(80) filename
      LOGICAL, DIMENSION(:,:), ALLOCATABLE :: EdgeFilter
      LOGICAL doWrite
C .. HPF dirctives to distribute the arrays ..
! An abstract processor array procs is declared, whose size NUMBER_OF_PROCESSORS()
! is declared at run time.  A 1D template is defined and distrubed by
! blocks onto the processor array.  The 1D arrays are simply aligned with the
! template, insuring that their distribution is the same as the template's.
! The 2D arrays have their first dimension collapsed (see the *),
! and the second dimension  is
! aligned with the template.

!HPF$ PROCESSORS             :: procs(NUMBER_OF_PROCESSORS())
!HPF$ TEMPLATE,DIMENSION(NN) :: tplate
!HPF$ DISTRIBUTE (BLOCK) ONTO procs :: tplate
!HPF$ ALIGN (:) WITH tplate(:) :: COStheta,SINtheta
!HPF$ ALIGN (*,:) WITH tplate(:) :: Hough,CountBase
!HPF$ ALIGN (*,:) WITH tplate(:) :: iRho,iRhoInd2
!HPF$ ALIGN (*,:) WITH tplate(:) :: countMask
C Begin Executable
      tcomp = 0.0
      tcomms = 0.0
C Default values.
      EsizeX = 204
      EsizeY = 124
      doWrite = .TRUE.
C Read in image filename.
      filename = 'image.pbm'
C Read in the Edge image.
      OPEN(UNIT=FILEIN,FILE=filename,STATUS='UNKNOWN')
      READ(UNIT=FILEIN,FMT='(A2)') magic
      READ(UNIT=FILEIN,FMT=*)
      READ(UNIT=FILEIN,FMT=*) EsizeX, EsizeY
      ALLOCATE( Edge(EsizeX,EsizeY) )
      ALLOCATE( XX(EsizeX,EsizeY) )
      ALLOCATE( YY(EsizeX,EsizeY) )
      ALLOCATE( EdgeFilter(EsizeX,EsizeY) )
      WRITE(UNIT=STDOUT,FMT='(''Reading image '',I4,'' x '',I4)')
     1     EsizeX,EsizeY
      READ(UNIT=FILEIN,FMT=*) ( (Edge(i,j), i=1,EsizeX),j=1,EsizeY )
      CLOSE(UNIT=FILEIN)
      WRITE(UNIT=STDOUT,FMT='(''Starting transformation'')')
      NedgePixels = SUM( Edge )
      ALLOCATE( XXpacked(NedgePixels) )
      ALLOCATE( YYpacked(NedgePixels) )
      ALLOCATE( iRho(NedgePixels,NN) )
      ALLOCATE( iRhoInd2(NedgePixels,NN) )
C Start timer
      CALL timer(ti, 0.0)
      Hough = 0
C Set up XX() and YY() arrays, equal to their index numbers.
C Then mask out elements where the edge pixels equal zero.
      EdgeFilter = .FALSE.
      WHERE( Edge .EQ. 1 ) EdgeFilter = .TRUE.
      FORALL( i=1:EsizeX,j=1:EsizeY ) XX(i,j) = i - 1
      XXpacked = PACK( ARRAY=XX,MASK=EdgeFilter )
      FORALL( i=1:EsizeX,j=1:EsizeY ) YY(i,j) = j - 1
      YYpacked = PACK( ARRAY=YY,MASK=EdgeFilter )
      DEALLOCATE( EdgeFilter )
      DEALLOCATE( YY )
      DEALLOCATE( XX )
      DEALLOCATE( Edge )
C Set up COStheta and SINtheta coefficient arrays.
      dTheta = PId2 / (NN - 1)
      dRho   = (NN - 1) /
     1         SQRT( REAL(EsizeX*EsizeX) + REAL(EsizeY*EsizeY) )
      FORALL(k=1:NN) COStheta(k) = COS( (k-1)*dTheta )
      FORALL(k=1:NN) SINtheta(k) = SIN( (k-1)*dTheta )
C Set up iRhoInd2 array, to be used in COUNT_SCATTER() operation.
      FORALL( k=1:NN ) iRhoInd2(:,k) = k
C Parallel loop over `theta' direction.
      FORALL( k=1:NN )
         iRho(:,k) = dRho*( XXpacked(:)*COStheta(k) +
     1                      YYpacked(:)*SINtheta(k) ) + 1.5
      END FORALL
C Set up tempoarary arrays for COUNT operation.
      ALLOCATE( CountMask( NedgePixels,NN ) )
      ALLOCATE( CountBase( NN,NN ) )
      CountMask = .TRUE.
      CountBase = 0
C Count up the occurences of iRho into Hough array.
C PGHPF on Cray requires `ARRAY=' *not* to be present.  04/Feb/97.
      Hough = COUNT_SCATTER( CountMask, BASE=CountBase,
     1                       INDX1=iRho, INDX2=iRhoInd2 )
C Stop timer
      CALL timer(tf, ti)
C Adjust the timer.  It would appear from the profiler that no time
C was spent on communications...
      tcomms = 0.0
      tcomp  = tf
      IF( doWrite )THEN
         WRITE(UNIT=STDOUT,FMT='(''Writing Transform image'')')
         OPEN(UNIT=FILEOUT,FILE='hough.pgm',STATUS='UNKNOWN')
         WRITE(UNIT=FILEOUT,FMT='(A)') 'P2'
         WRITE(UNIT=FILEOUT,FMT='(A)')
     1      '# Created by Hough transformation program.  NPAC.'
         WRITE(UNIT=FILEOUT,FMT='(2I6)') NN,NN
         WRITE(UNIT=FILEOUT,FMT='(I6)') MAXVAL( Hough )
         WRITE(UNIT=FILEOUT,FMT=*) ((Hough(i,j), i=1,NN),j=1,NN)
         CLOSE(UNIT=FILEOUT)
      END IF
      WRITE(UNIT=*,FMT=6004) NUMBER_OF_PROCESSORS(),NN,
     1                       tcomms,tcomp,(tf-tcomms-tcomp),tf
      DEALLOCATE( CountBase )
      DEALLOCATE( CountMask )
      DEALLOCATE( iRhoInd2 )
      DEALLOCATE( iRho )
      DEALLOCATE( YYpacked )
      DEALLOCATE( XXpacked )
      STOP
 1001 FORMAT('Status: iTime = ',I6,' Error = ',F9.4)
 1002 FORMAT(12F9.3)
 6004 FORMAT('Number of Processors = ',I4/
     1       'Problem size = ',I5/
     2       'Communications = ',F9.3/
     3       'Compute        = ',F9.3/
     4       'Others         = ',F9.3/
     5       'Total time     = ',F9.3)
      END
C============================================================


       subroutine timer(return_time, initial_time)
       real, intent(IN)  :: initial_time
       real, intent(OUT)  :: return_time
       integer finish,rate
        CALL system_clock(COUNT=finish,COUNT_RATE=rate)
        return_time = FLOAT(finish) / FLOAT (rate) -initial_time
        return
       end
