!  A.1 NAS EP Benchmark - Tabulation of Random
!  Numbers

! This is the NPAC version of the NAS Embarrassingly Parallel benchmarks code,
! and generates random numbers with a Gaussian probability distribution, based on the
! Box-Muller algorithm. The timing information output lists time spent
! executing statements with communication required, without communication, and
! others.  The no-communciation times should scale nicely, but those with
! communication obviously will not.

      PROGRAM ep
      IMPLICIT NONE
      INTEGER, PARAMETER          :: M = 20, MK = 4
      INTEGER, PARAMETER          :: MM = M-MK
      INTEGER, PARAMETER          :: NK = 2**MK
      INTEGER, PARAMETER          :: NN = 2**MM
      INTEGER, PARAMETER          :: NQ = 10
C M  = log2 number of [0,1] uniform random pairs to create.
C MK = log2 size of each batch of [0,1] uniform pairs to create.
C      Where NK = 2**MK.
C MM = log2 number of independent batches to create, divisible with
C      the number of available processors.
C      Where NN = 2**MM.
C NQ = Number of bins with which to gather statistics on the Gaussian
C      deviates.
      DOUBLE PRECISION, DIMENSION(NN) :: X,Y
      DOUBLE PRECISION, DIMENSION(NN) :: X1,X2,T2,T3,T4,GC
      INTEGER, DIMENSION(NN)          :: KK
      INTEGER, DIMENSION(0:NQ,NN) :: QQ
      INTEGER, DIMENSION(0:NQ)    :: Q
      INTEGER I,J,NGCI
      INTEGER iStep
      DOUBLE PRECISION NGC
      DOUBLE PRECISION, PARAMETER :: A = 1220703125.D0
      DOUBLE PRECISION, PARAMETER :: S =  271828183.D0
      REAL :: ti,tf
      REAL :: tstart,tend
      REAL :: tcomms, tcomp
      REAL :: randComp
!
! The following HPF statements first define a one-dimensional
! processor array procs,which has a size  NUMBER_OF_PROCESSORS(),
! the number specified at runtime by the user. 
! A template of size NN is defined, then distributed in a BLOCK
! fashion onto the processor array.  The other one-dimensional
! arrays are aligned with the template, thus insuring they are
! distributed in the same fashion as the template. 
! The two dimensional array QQ has its second dimension aligned
! with the template TTT, while its first dimension is collapsed (*,:).
! Thus the second dimension is distributed in a block fashion like
! TTT.

!HPF$ PROCESSORS procs( NUMBER_OF_PROCESSORS() )
!HPF$ TEMPLATE TTT(NN)
!HPF$ DISTRIBUTE TTT(BLOCK) ONTO procs
!HPF$ ALIGN (:) WITH TTT(:) :: X,Y,X1,X2,T2,T3,T4,GC,KK
!HPF$ ALIGN QQ(*,:) WITH TTT(:)


      INTERFACE
        SUBROUTINE random_seed(s, a, x, MK, NN)
        DOUBLE PRECISION, DIMENSION(:), INTENT(OUT) :: x
!HPF$   DISTRIBUTE *(BLOCK) :: x
        DOUBLE PRECISION, INTENT(IN)             :: a,s
        INTEGER, INTENT(IN)                      :: MK,NN
        END SUBROUTINE RANDOM_SEED
        SUBROUTINE randlc (x, a, y, NN, myComp)
        DOUBLE PRECISION, DIMENSION(:), INTENT(INOUT)  :: x
        DOUBLE PRECISION, DIMENSION(:), INTENT(OUT)    :: y
        DOUBLE PRECISION, INTENT(IN)                :: a
        INTEGER, INTENT(IN)                         :: NN
        REAL, INTENT(INOUT)                         :: myComp
!HPF$   DISTRIBUTE *(BLOCK)  :: x,y
        END SUBROUTINE randlc
      END INTERFACE
      INTERFACE
         SUBROUTINE timer(return_time, initial_time)
         REAL, INTENT(IN)  :: initial_time
         REAL, INTENT(OUT) :: return_time
         END SUBROUTINE timer
      END INTERFACE
C Begin Executable
      tcomms = 0.0
      tcomp  = 0.0
      randComp = 0.0
      iStep = NK / 10
      IF( iStep .EQ. 0 ) iStep = 1
      X  = 0.0
      Y  = 0.0
      X1 = 0.0
      X2 = 0.0
      T2 = 0.0
      T3 = 0.0
      T4 = 0.0
      GC = 0.0
      KK = 0
      QQ = 0
      Q  = 0
C start_timer
      CALL timer(ti, 0.0)
      CALL random_seed(S,A,X,MK,NN)
      QQ=0
      DO I = 1, NK
         IF( MOD(i,iStep) .EQ. 0 )THEN
            WRITE(UNIT=*,FMT='(''Status i = '',I8)') i
         END IF
C Compute Gaussian deviates by acceptance-rejection method and
C tally counts in concentric square annuli.
        CALL randlc (x, a, y, NN, randComp)
        CALL timer(tstart,0.0)
        X1 = 2.D0 * Y - 1.D0
        CALL timer(tend,tstart)
        tcomp = tcomp + tend
        CALL randlc (x, a, y, NN, randComp)
        CALL timer(tstart,0.0)
        X2 = 2.D0 * Y - 1.D0
        Y = X1 ** 2 + X2 ** 2
        WHERE (Y .LE. 1.D0)
          GC = GC + 1.D0
          T2 = SQRT (-2.D0 * LOG (Y) / Y)
          T3 = ABS (X1 * T2)
          T4 = ABS (X2 * T2)
          KK = MAX (T3, T4)
        ELSEWHERE
          KK = NQ
          Y = -10.0D0
        ENDWHERE
        DO J = 0, NQ - 1
          WHERE (KK .EQ. J) QQ(J,:) = QQ(J,:) + 1
        ENDDO
        CALL timer(tend,tstart)
        tcomp = tcomp + tend
      ENDDO
      CALL timer(tstart,0.0)
      Q   = SUM(QQ,DIM=2)
      NGC = SUM(GC)
      CALL timer(tend,tstart)
      tcomms = tcomms + tend
C Stop timer.
      CALL timer(tf,ti)
      tcomp = tcomp + randComp
C This is done so the WRITE statement works
      NGCI = INT(NGC)
      PRINT *,'time taken is ',tf,'seconds'
      WRITE(6,1001) M,NGCI,(I,Q(I),I=0,NQ-1)
      WRITE(UNIT=*,FMT=6004) NUMBER_OF_PROCESSORS(),NN,
     1                       tcomms,tcomp,(tf-tcomms-tcomp),tf
      STOP
 1001 FORMAT('BENCHMARK 1 RESULTS:' // 'N = 2^',I8/
     &   'NO. OF GAUSSIAN PAIRS =',I15/'COUNTS:'/(I3,I15))
 2001 FORMAT(12E12.3)
 6004 FORMAT('Number of Processors = ',I4/
     1       'Problem size = ',I8/
     2       'Communications = ',F9.3/
     3       'Compute        = ',F9.3/
     4       'Others         = ',F9.3/
     5       'Total time     = ',F9.3)
      END
C======================================================================
      SUBROUTINE random_seed(s, a, x, MK,NN)
C _____________________________________________________________________
C|                                                                     |
C| Routine to generate array x(NN) of seeds for parallel random number |
C| generator.                                                          |
C|                                                                     |
C| As with RANDLC, compute R23 = 2 ^ -23, R46 = 2 ^ -46,               |
C| T23 = 2 ^ 23, and T46 = 2 ^ 46.  These are computed in loops, rather|
C| than by merely using the ** operator, in order to insure that the   |
C| results are exact on all systems.  This code assumes that 0.5D0 is  |
C| represented exactly.                                                |
C|                                                                     |
C| H W Yau.  15th of August, 1996.                                     |
C| Northeast Parallel Architectures Center.                            |
C| Syracuse University.                                                |
C|_____________________________________________________________________|
C| Edit record:                                                        |
C| 15/Aug/1996: Corrected comments.                                    |
C|              Corrected generation of the seeds x(NN).               |
C|_____________________________________________________________________|
      IMPLICIT NONE
      DOUBLE PRECISION, DIMENSION(:), INTENT(OUT) :: x
      DOUBLE PRECISION, INTENT(IN)             :: a, s
      INTEGER, INTENT(IN)                      :: MK,NN
! The lone HPF statement in this subroutine simply
! tells the compiler to leave the incoming distribution of array
! x as it is in the calling program.  The * in *(BLOCK) denotes this.

!HPF$ DISTRIBUTE *(BLOCK) :: x
      DOUBLE PRECISION :: t1,t2,zz1,zz2,z
      DOUBLE PRECISION R23, R46, T23, T46, an, an1, an2
      DOUBLE PRECISION, DIMENSION(NN) :: zz
      INTEGER I
C Begin Executable
      R23 = 1.D0
      R46 = 1.D0
      T23 = 1.D0
      T46 = 1.D0
      DO  I = 1, 23
         R23 = 0.5D0 * R23
         T23 = 2.0D0 * T23
      ENDDO
      DO I = 1, 46
         R46 = 0.5D0 * R46
         T46 = 2.0D0 * T46
      ENDDO
C Calculate: an = a^MK mod 2^46.
      an = a
      DO i = 1, MK+1
C Perform (an * an) MOD 2^46.  First, split `an' into two parts such that
C an = 2^23 * an1 + an2.
         an1 = AINT( R23 * an )
         an2 = an - T23 * an1
C Compute z = an1*an2 + an2*an1 (mod 2^23), and then
C an = 2^23 * z + an2*an2 (mod 2^46).
         t1 = an1 * an2 + an2 * an1
         z  = t1 - T23 * AINT( R23 * t1 )
         t2 = T23 * z + an2 * an2
         an = t2 - T46 * AINT( R46 * t2 )
      END DO
C Now compute the seeds, into a non-distributed array zz().  First,
C split `an' into two parts such that
C an = 2^23 * an1 + an2.
      an1 = AINT( R23 * an )
      an2 = an - T23 * an1
      zz(1) = s
      DO i = 2,NN
C Break z(i-1) into two parts, such that z(i-1) = 2^23 * zz1 + zz2.
         zz1 = AINT( R23 * zz(i-1) )
         zz2 = zz(i-1) - T23 * zz1
C Compute z = an1*zz2 + an2*zz2 (mod 2^23), and then
C     zz(i) = 2^23*z + an2*zz2 (mod 2^46).
         t1    = an1 * zz2 + an2 * zz1
         z     = t1 - T23 * AINT( R23 * t1 )
         t2    = T23 * z + an2 * zz2
         zz(i) = t2 - T46 * AINT( R46 * t2 )
      END DO
C Copy into distributed array.
      x = zz
      RETURN
      END
C======================================================================
      SUBROUTINE RANDLC (X, A, Y, NN, myComp)
      IMPLICIT NONE
      DOUBLE PRECISION, DIMENSION(:), INTENT(INOUT)  :: X
      DOUBLE PRECISION, DIMENSION(:), INTENT(OUT)    :: Y
      DOUBLE PRECISION, INTENT(IN):: A
      INTEGER, INTENT(IN)         :: NN
      REAL, INTENT(INOUT)         :: myComp
      DOUBLE PRECISION, DIMENSION(SIZE(X)) :: A1,A2,T1,T2,T3,T4,X1,X2,Z
      DOUBLE PRECISION R23, R46, T23, T46
      REAL :: tstart, tend
      INTEGER I,KS
      SAVE KS, R23, R46, T23, T46
      DATA KS /0/
! The first HPF statement tells the compiler to leave the
! distribution of x and y as it is in the calling routine.
! Then the local arrays are aligned with these, so that
! the distribution is the same.

!HPF$ DISTRIBUTE  *(BLOCK) :: x,y
!HPF$ ALIGN (:) WITH x(:) :: A1,A2,T1,T2,T3,T4,X1,X2,Z
C A is constant
C given array X(NN) of seeds
C generate array Y(NN) of random numbers
C If this is the first call to RANDLC, compute R23 = 2 ^ -23, R46 = 2 ^ -46,
C T23 = 2 ^ 23, and T46 = 2 ^ 46.  These are computed in loops, rather than
C by merely using the ** operator, in order to insure that the results are
C exact on all systems.  This code assumes that 0.5D0 is represented exactly.
C Begin Executable
      CALL timer(tstart,0.0)
      IF (KS .EQ. 0) THEN
         R23 = 1.D0
         R46 = 1.D0
         T23 = 1.D0
         T46 = 1.D0
         DO  I = 1, 23
            R23 = 0.5D0 * R23
            T23 = 2.D0 * T23
         ENDDO
         DO  I = 1, 46
            R46 = 0.5D0 * R46
            T46 = 2.D0 * T46
         ENDDO
         KS = 1
      ENDIF
C Break A into two parts such that A = 2^23 * A1 + A2.
      T1 = R23 * A
      A1 = AINT (T1)
      A2 = A - T23 * A1
C Break X into two parts such that X = 2^23 * X1 + X2, compute
C Z = A1 * X2 + A2 * X1  (mod 2^23, and then
C X = 2^23 * Z + A2 * X2  (mod 2^46).
      T1 = R23 * X
      X1 = AINT (T1)
      X2 = X - T23 * X1
      T1 = A1 * X2 + A2 * X1
      T2 = AINT (R23 * T1)
      Z = T1 - T23 * T2
      T3 = T23 * Z + A2 * X2
      T4 = AINT (R46 * T3)
      X = T3 - T46 * T4
      Y = R46 * X
      CALL timer(tend,tstart)
      myComp = myComp + tend
      RETURN
      END
C======================================================================


       subroutine timer(return_time, initial_time)
       real, intent(IN)  :: initial_time
       real, intent(OUT)  :: return_time
       integer finish,rate
        CALL system_clock(COUNT=finish,COUNT_RATE=rate)
        return_time = FLOAT(finish) / FLOAT (rate) -initial_time
        return
       end
