!  A.4 Gaussian Elimination

! This code performs factorization of a random matrix, through
! Gaussian elimination with partial pivoting. 
! The timing information output lists time spent
! executing statements with communication required, without communication, and
! others.  The no-communciation times should scale nicely, but those with
! communication obviously will not.

      PROGRAM gauss
      IMPLICIT NONE
      INTEGER, PARAMETER :: Nrows = 255, Ncols = 256
C Define interface for subroutine timer.
      INTERFACE
        SUBROUTINE timer(return_time, initial_time)
          REAL, INTENT(IN)  :: initial_time
          REAL, INTENT(OUT) :: return_time
        END SUBROUTINE timer
      END INTERFACE
C Mask, for setting which rows in which to search for next pivot.
      INTEGER, DIMENSION(Nrows) :: pivotMask
      DOUBLE PRECISION, DIMENSION(Nrows,Ncols) :: a,pivotColArray
      DOUBLE PRECISION, DIMENSION(Nrows) :: tmpCol,x
      DOUBLE PRECISION, DIMENSION(Ncols) :: pivotRow
C Matrix row, column, and pivot-stage indices.
      INTEGER :: i,j,k
      INTEGER :: ipivot
      INTEGER, DIMENSION(1) :: ipivotArray
      DOUBLE PRECISION :: pivot,tmp
      REAL :: ti,tf
      REAL :: tstart,tend
      REAL :: tcomms, tcomp
C HPF Directives

! The HPF statements create a 1D abstract processor array procs of
! size NUMBER_OF_PROCESSORS(), which is declared at run time.  
! A template is created and distributed in a CYCLIC fashion,
! which makes sense for this problem because it insures good load
! balancing.  A Block distribution would cause load balancing
! problems.  The 2D arrays are collapsed along the first dimension
! and distributed in the same CYCLIC fashion as the template
! along their second dimension.  The align statements insure this fact.

!HPF$ PROCESSORS procs(NUMBER_OF_PROCESSORS())
!HPF$ TEMPLATE temp(Ncols)
!HPF$ DISTRIBUTE temp(CYCLIC)
!HPF$ ALIGN a(*,j)             WITH temp(j)
!HPF$ ALIGN pivotColArray(*,j) WITH temp(j)
!HPF$ ALIGN pivotRow(j)        WITH temp(j)
C Executable Code.
      tcomp = 0.0
      tcomms = 0.0
C Generate random matrix.
      CALL RANDOM_NUMBER( a )
C Calculate appropriate RHS into Ncols-th column of a(), such
C that the solution should equal to 1.0,2.0,3.0,...Nrows.
      FORALL( i=1:Nrows ) x(i) = REAL(i)
      tmpCol = 0.0D0
      DO i=1,Nrows
        FORALL( j=1:Nrows ) tmpCol(j) = a(i,j)*x(j)
        a(i,Ncols) = SUM( tmpCol )
      END DO
      x = 0.0D0
C Start timer
      CALL timer(ti, 0.0)
      pivotMask = 1
      DO k=1,Nrows-1
C Use tmpCol as temporary column to find the pivot.
        CALL timer(tstart,0.0)
        tmpCol = ABS( a(:,k) )
        CALL timer(tend,tstart)
        tcomms = tcomms + tend
C Find the pivot in the k-th column, within the rows k:Nrows.
        ipivotArray = MAXLOC( tmpCol,MASK=pivotMask.EQ.1 )
        ipivot = ipivotArray(1)
        IF( ipivot .NE. k )THEN
C Swap rows, using pivotRow as a temporary storage.
          pivotRow(:) = a(ipivot,:)
          a(ipivot,:) = a(k,:)
          a(k,:) = pivotRow(:)
        ELSE
          pivotRow(:) = a(k,:)
        END IF
C Define the pivot, and the (rescaled) pivotColArray.  This is the
C pivot column replicated across all columns.
        CALL timer(tstart,0.0)
        pivot = a(k,k)
        FORALL( i=k+1:Nrows,j=1:Ncols )
     $         pivotColArray(i,j) = a(i,k)/pivot
        CALL timer(tend,tstart)
        tcomms = tcomms + tend
C Calculate upper triangular factor, by updating matrix using the
C pivot, pivotRow and pivotColarray.
        FORALL( i=k+1:Nrows,j=k+1:Ncols )
     $          a(i,j) = a(i,j) - pivotColArray(i,j)*pivotRow(j)
C Save pivot column for the lower triangular factor.
        a(k+1:Nrows,k) = pivotColArray(k+1:Nrows,k)
        pivotMask(k) = 0
      END DO
C Now do backwards substitution.
      DO i=Nrows,1,-1
        CALL timer(tstart,0.0)
        tmp = a(i,Ncols)
        CALL timer(tend,tstart)
        tcomms = tcomms + tend
        tmpCol = 0.0D0
        CALL timer(tstart,0.0)
        FORALL( j=1:Nrows,j.GT.i ) tmpCol(j) = a(i,j)*x(j)
        CALL timer(tend,tstart)
        tcomms = tcomms + tend
        tmp = tmp - SUM( tmpCol )
        CALL timer(tstart,0.0)
        x(i) = tmp/a(i,i)
        CALL timer(tend,tstart)
        tcomms = tcomms + tend
      END DO
C Stop timer
      CALL timer(tf, ti)
      tcomp = tf - ti - tcomms
      WRITE(UNIT=*,FMT=*) 'Solution is:'
      WRITE(UNIT=*,FMT=101) (x(i),i=1,Nrows)
      WRITE(UNIT=*,FMT=6004) NUMBER_OF_PROCESSORS(),Ncols,
     1                       tcomms,tcomp,(tf-tcomms-tcomp),tf
      STOP
 101  FORMAT(10F9.3)
 6004 FORMAT('Number of Processors = ',I4/
     1       'Problem size = ',I5/
     2       'Communications = ',F9.3/
     3       'Compute        = ',F9.3/
     4       'Others         = ',F9.3/
     5       'Total time     = ',F9.3)
      END
C======================================================================
	 	subroutine timer(return_time, initial_time)
       real, intent(IN)  :: initial_time
       real, intent(OUT)  :: return_time
       integer finish,rate
        CALL system_clock(COUNT=finish,COUNT_RATE=rate)
        return_time = FLOAT(finish) / FLOAT (rate) -initial_time
        return
       end
