! Hopfield Neural Network

! The Hopfield neural network is a simple associative
!memory model with the ability to store and later retrieve a
!given set of images. In the version below, the code
!updates the state of each neuron (referred to as `sites' in
!the code) via a stochastic parallel update, with the
!probability of update a user-tunable parameter. Much of
!the computation is matrix calculations, which makes this
!code amenable to HPF's data parallelism execution model.
! The timing information output lists time taken   
! executing statements with communication required, without communication, and
! others.  The no-communciation times should scale nicely, but those with
! communication obviously will not.

      PROGRAM main
      IMPLICIT NONE
      INTERFACE
        SUBROUTINE timer(return_time, initial_time)
          REAL,INTENT(IN)  :: initial_time
          REAL,INTENT(OUT) :: return_time
        END SUBROUTINE timer
      END INTERFACE
      INTEGER,PARAMETER :: STDOUT        = 6
      INTEGER,PARAMETER :: MaxIterations = 100
      REAL,PARAMETER    :: tol           = 0.99
      REAL,PARAMETER    :: updateFrac    = 1.0
      REAL :: ti,tf
      REAL :: tstart,tend
      REAL :: tcomms, tcomp
      REAL,DIMENSION(:),ALLOCATABLE   :: RealRandom
      REAL,DIMENSION(:),ALLOCATABLE   :: LocalField
      REAL,DIMENSION(:),ALLOCATABLE   :: ExternalField
      REAL,DIMENSION(:),ALLOCATABLE   :: overlap
      REAL,DIMENSION(:,:),ALLOCATABLE :: weights
      INTEGER,DIMENSION(:),ALLOCATABLE   :: sites,sites0
      INTEGER,DIMENSION(:,:),ALLOCATABLE :: patterns
      REAL :: MaxOverlap
      REAL :: NoiseFrac
      REAL :: SelfOverlap
      INTEGER :: iarg
      INTEGER :: i,iter,n,Npatterns
      CHARACTER*(128) :: argString
C ... HPF directives ...
! These HPF statements first define the 1D abstract processor array,
! whose size is specified at runtime.  The LocalField array is
! distributed by blocks on the processor array.  The 1D array External Field
! is aligned with LocalField, as is the  second dimension of the 2D array
! weights. Patterns has its first dimension aligned with the elements of
! LocalField.  The elements of the aligned dimension are distributed in the
! same manner as the alignee.  The * dimension is collapsed.
!HPF$ PROCESSORS :: procs( NUMBER_OF_PROCESSORS() )
!HPF$ DISTRIBUTE(BLOCK) ONTO procs :: LocalField
!HPF$ ALIGN ExternalField(:) WITH LocalField(:)
!HPF$ ALIGN weights(*,:)     WITH LocalField(:)
!HPF$ ALIGN patterns(:,*)    WITH LocalField(:)
C Begin Executable
C ... Default settings (designed to loop to MaxIterations)
      tcomms = 0.0
      tcomp  = 0.0
      iarg = 0
      NoiseFrac = 0.40
      n = 256
      Npatterns = INT( (n * 0.14) + 0.5)
C PGI HPF does not implement IARGC().
      IF( iarg .GE. 1 )THEN
         CALL getarg(1,argString)
         READ(UNIT=argString,FMT=*) NoiseFrac
         WRITE(UNIT=STDOUT,FMT='(''Setting Starting Noise  = '',F12.3)')
     1         NoiseFrac
      END IF
      IF( iarg .GE. 2 )THEN
         CALL getarg(2,argString)
         READ(UNIT=argString,FMT=*) n
         WRITE(UNIT=STDOUT,FMT='(''Setting array size = '',I6)') n
         Npatterns = INT( (n * 0.14) + 0.5)
      END IF
      IF( iarg .GE. 3 )THEN
         CALL getarg(3,argString)
         READ(UNIT=argString,FMT=*) Npatterns
         WRITE(UNIT=STDOUT,FMT='(''Setting Number of patterns = '',I6)')
     1         Npatterns
      END IF
      ALLOCATE( sites(n) )
      ALLOCATE( sites0(n) )
      ALLOCATE( RealRandom(n) )
      ALLOCATE( LocalField(n) )
      ALLOCATE( ExternalField(n) )
      ALLOCATE( overlap(Npatterns) )
      ALLOCATE( weights(n,n) )
      ALLOCATE( patterns(n,Npatterns) )
C Start timer
      CALL timer(ti, 0.0)
      CALL timer(tstart,0.0)
C ... Define patterns (-1,+1) with which the network will learn.
      DO i=1,Npatterns
         CALL RANDOM_NUMBER( RealRandom )
         WHERE( RealRandom .GE. 0.5 )
            patterns(:,i) = 1.0
         ELSEWHERE
            patterns(:,i) = -1.0
         END WHERE
      END DO
C ... Hebbian learning rule for the weights ...
      weights = 0.0
      DO i=1,Npatterns
         sites = patterns(:,i)
         weights = weights + MATMUL( RESHAPE(sites,(/n,1/)),
     1                               RESHAPE(sites,(/1,n/)) )
      END DO
      weights = weights / REAL( n )
C ... Define external field ...
      ExternalField = 0.01
C ... Pick the last pattern stored as the one to retrieve.
      sites = patterns(:,Npatterns)
C ... Add some noise.
      CALL RANDOM_NUMBER( RealRandom )
      WHERE( RealRandom .LE. noiseFrac )
         sites = -sites
      END WHERE
      sites0 = sites
      overlap = RESHAPE( MATMUL( RESHAPE(sites,(/1,n/)),patterns ),
     1                   (/Npatterns/) )
      overlap = overlap / REAL( n )
      MaxOverlap = MAXVAL( overlap )
      iter = 0
      SelfOverlap = 0.0
      WRITE(UNIT=STDOUT,FMT=102) iter,SelfOverlap
      CALL timer(tend, tstart)
      tcomp = tcomp + tend
      DO WHILE( (MaxOverlap .LE. tol).AND.(iter.NE.MaxIterations) )
         iter = iter + 1
C ... Calculate the local field at each site.
         CALL timer(tstart, 0.0)
         LocalField = RESHAPE( MATMUL( RESHAPE(sites,(/1,n/)),weights ),
     1                         (/n/) ) + ExternalField
         CALL timer(tend, tstart)
         tcomms = tcomms + tend
         CALL timer(tstart, 0.0)
C ... Noisy parallel update dynamics, using sign() function.
         CALL RANDOM_NUMBER( RealRandom )
         WHERE( (LocalField.GE.0.0).AND.(RealRandom.LE.updateFrac) )
            sites = 1
         END WHERE
         WHERE( (LocalField.LT.0.0).AND.(RealRandom.LE.updateFrac) )
            sites = -1
         END WHERE
         CALL timer(tend,tstart)
         tcomp = tcomp + tend
         CALL timer(tstart, 0.0)
         overlap = RESHAPE( MATMUL( RESHAPE(sites,(/1,n/)),patterns ),
     1                      (/Npatterns/) )
         CALL timer(tend, tstart)
         tcomms = tcomms + tend
         CALL timer(tstart, 0.0)
         overlap = overlap / REAL( n )
         MaxOverlap = MAXVAL( overlap )
         SelfOverlap = DOT_PRODUCT( sites,sites0 ) / REAL(n)
         sites0 = sites
         CALL timer(tend, tstart)
         tcomp = tcomp + tend
         WRITE(UNIT=STDOUT,FMT=102) iter,SelfOverlap
      END DO
C Stop timer
      CALL timer(tf, ti)
      WRITE(UNIT=STDOUT,FMT=101) MAXLOC( overlap ), MaxOverlap
      WRITE(UNIT=*,FMT=6004) NUMBER_OF_PROCESSORS(),n,
     1                       tcomms,tcomp,(tf-tcomms-tcomp),tf
 6004 FORMAT('Number of Processors = ',I4/
     1       'Problem size = ',I5/
     2       'Communications = ',G9.3/
     3       'Compute        = ',G9.3/
     4       'Others         = ',G9.3/
     5       'Total time     = ',G9.3)
      STOP
 101  FORMAT('Pattern found is number ',I4,' with overlap = ',F12.3)
 102  FORMAT('Progress',I6,':',F12.3)
 103  FORMAT(12F9.3)
      END

       subroutine timer(return_time, initial_time)
       real, intent(IN)  :: initial_time
       real, intent(OUT)  :: return_time
       integer finish,rate
        CALL system_clock(COUNT=finish,COUNT_RATE=rate)
        return_time = FLOAT(finish) / FLOAT (rate) -initial_time
        return
       end

