!A.6 2-Dimensional Potts Model

! This code sets up a 2-D Q-State Potts model, and evolves it using the classic
!Metropolis heat-bath algorithm. The timing information output lists time spent
! executing statements with communication required, without communication, and
! others.  The no-communciation times should scale nicely, but those with
! communication obviously will not.


      PROGRAM potts2d
      IMPLICIT NONE
      integer, parameter :: m = 64 ! Lattice dimensions
      integer, parameter :: n = 64
      real ti,start, tf,tend,tcom
      real correct1,correct2
      real tcalc

! These HPF statements first creat a one dimensional abstract processor
! array PROCS, of size NUMBER_OF_PROCESSORS(), determined at runtime.
! A two dimensional template is declared, and then it is distributed
! in blocks along its second dimension.  The first dimension is
! collapsed (note the * in (*,BLOCK).
!HPF$ PROCESSORS PROCS(NUMBER_OF_PROCESSORS())
!HPF$ TEMPLATE, DIMENSION(m,n) :: rectangle
!HPF$ DISTRIBUTE (*,BLOCK) ONTO PROCS :: rectangle

      interface
         subroutine timer(return_time, initial_time)
         real, intent(in)  :: initial_time
         real, intent(out) :: return_time
         end subroutine timer
         subroutine setup_random(iseed)
         integer, intent(IN) :: iseed
         end subroutine setup_random
      end interface

C     Number of allowable States for Degrees of Freedom
      integer, parameter :: q = 4
C     Coupling in kT units, -ve for ferromagnetic
      real, parameter :: cup = -4.0
C     The lattice of DOF's
      integer, dimension( 1:m, 1:n ) :: latt
! The following HPF statements align various two-dimensional arrays
! with the template , thus insuring that they are distributed in
! the same manner as the template, and each other.

!HPF$ ALIGN (:,:) WITH RECTANGLE :: latt
C     Trial choices for DOF's
      integer, dimension( 1:m, 1:n ) :: itmp
!HPF$ ALIGN (:,:) WITH RECTANGLE :: itmp
C     TMP for RNG and Probability
      real, dimension( 1:m, 1:n ) :: tmp, p
!HPF$ ALIGN (:,:) WITH RECTANGLE :: tmp
!HPF$ ALIGN (:,:) WITH RECTANGLE :: p
C     mask used in ensuring detailed balance prevails
      logical, dimension( 1:m, 1:n ) :: mask
C     !HPF$ ALIGN mask WITH RECTANGLE
C     number of measurements
      integer :: nmeasure = 20
C     number of sweeps per measurement
      integer :: niter = 10
      integer  imeasure,iter
      integer  i0,i,j,j0
C Begin Executable
      tcom = 0.0
      tcalc = 0.0
      latt = 0
      itmp = 0
      tmp = 0.0
      p = 0.0
      mask = .FALSE.
C     BEGIN TIMER
      call timer(start,0.0)
      call timer(ti,0.0)
      call setup_random( 12345 )
C     initialise between 0 and q-1 inclusive
      call random_number( tmp )
      latt = int( tmp * q )
      call timer(tcalc,ti)
      do imeasure=1,nmeasure
C     Measurements:
         call timer(ti,0.0)
! Note the use of the cshift operation, cshift(latt,  1, 1 ) which
! takes an integer specifying the dimension to shift along (1 or
! 2 in this case) and the offset (either 1 or -1 here) and
! thus shifts the array latt.

C     Present bonds
         itmp =        merge( 1, 0, latt == cshift(latt,  1, 1 ) )
         itmp = itmp + merge( 1, 0, latt == cshift(latt, -1, 1 ) )
         call timer(tend,ti)
         tcalc=tcalc + tend
         call timer(ti,0.0)
         itmp = itmp + merge( 1, 0, latt == cshift(latt,  1, 2 ) )
         itmp = itmp + merge( 1, 0, latt == cshift(latt, -1, 2 ) )
         call timer(tend,ti)
         tcom= tcom + tend
         write(6,*) '# Bonds = ', sum( itmp )
         do iter=1,niter
            call timer(ti,0.0)
C     generate "new" values for the Potts variables
            call random_number(tmp)
C     using the Coddington recipe
            itmp = mod( (latt + int( tmp * q-1 ) + 1 ), q )
            call timer(tend,ti)
            tcalc=tcalc+tend
C     use four-colour ordering to preserve detailed-balance
            do i0=1,2
               do j0=1,2
                  call timer(ti,0.0)
C     have to get too many cos of this stupid non-function RNG
                  call random_number( tmp )
                  mask = .false.
                  forall(i=i0:m:2, j=j0:n:2 ) mask(i,j) = .true.
                  where (mask)
C     Present bonds
                     p =     merge( cup, 0.0,
     1                              latt == cshift(latt,  1, 1 ) )
                     p = p + merge( cup, 0.0,
     1                              latt == cshift(latt, -1, 1 ) )
                  end where
                  call timer(tend,ti)
                  tcalc=tcalc+tend
                  call timer(ti,0.0)
                  where (mask)
                     p = p + merge( cup, 0.0,
     1                              latt == cshift(latt,  1, 2 ) )
                     p = p + merge( cup, 0.0,
     1                              latt == cshift(latt, -1, 2 ) )
                  end where
                  call timer(tend,ti)
                  tcom=tcom+tend
                  call timer(ti,0.0)
                  where (mask)
C     future bonds
                     p = p - merge( cup, 0.0,
     1                              itmp == cshift(latt,  1, 1 ) )
C     if dof is flipped
                     p = p - merge( cup, 0.0,
     1                              itmp == cshift(latt, -1, 1 ) )
                  end where
                  call timer(tend,ti)
                  tcalc=tcalc+tend
                  call timer(ti,0.0)
                  where (mask)
                     p = p - merge( cup, 0.0,
     1                              itmp == cshift(latt,  1, 2 ) )
                     p = p - merge( cup, 0.0,
     1                              itmp == cshift(latt, -1, 2 ) )
                  end where
                  call timer(tend,ti)
                  tcom=tcom+tend
                  call timer(ti,0.0)
                  where (mask)
                     p = exp( p )
                     latt = merge( itmp, latt, tmp < p )
                  end where
                  call timer(tend,ti)
                  tcalc=tcalc+tend
               end do
            end do              ! end of checkerboard loop
         end do                 ! end of iter loop
      end do                    ! end of measure loop
      call timer(tf,start)
C     END TIMER
      write(6,6001) NUMBER_OF_PROCESSORS(),
     1     m,tcom,tcalc,(tf-tcom-tcalc),tf
 6001 FORMAT('Number of Processors =',I4/
     1     'Problem size =',I4/
     2     'Communications = ',F9.3/
     3     'Compute = ', F9.3/
     4     'Others =', F9.3/
     5     'Total time =',F9.3 )
C write out picture of configuration as a PPM file
      write(9,100)
 100  Format('P3' )
      write(9,101)
 101  Format('# 2d Potts Model')
      write(9,FMT='(2I6)') m, n
      write(9,FMT='(I6)') 255
      do i=1,m
         do j=1,n
            if( latt(i,j) == 0 )then
               write(9,'(3I4)',ADVANCE='NO')
     1                         255,0,0
            else if( latt(i,j) == 1 )then
               write(9,'(3I4)',ADVANCE='NO')
     1                         0,255,0
            else if( latt(i,j) == 2 )then
               write(9,'(3I4)',ADVANCE='NO')
     1                         0,0,255
            else if( latt(i,j) == 3 )then
               write(9,'(3I4)',ADVANCE='NO')
     1                         255,255,0
            else if( latt(i,j) == 4 )then
               write(9,'(3I4)',ADVANCE='NO')
     1                         255,0,255
            else if( latt(i,j) == 5 )then
               write(9,'(3I4)',ADVANCE='NO')
     1                         0,255,255
            else if( latt(i,j) == 6 )then
               write(9,'(3I4)',ADVANCE='NO')
     1                         255,255,255
            else
               write(9,'(3I4)',ADVANCE='NO')
     1                         0,0,0
            endif
C Insert line break, as required by Cray Fortran.
            if( MOD( ((i-1)*n)+j,5 ) .EQ. 0 ) write(9,FMT=*)
         enddo
      enddo
      end program
      subroutine setup_random(iseed)
C     put a new seed in the bloody awful system supplied RNG
      IMPLICIT NONE
      integer :: n,i,iseed
      integer, dimension(:), allocatable ::  seed
      call random_seed(size=n)
      allocate( seed(n) )
      forall(i=1:n) seed(i) = mod(i*4096+150889+iseed,714025 )
      call random_seed( put=seed(1:n) )
      deallocate( seed )
      end subroutine
      subroutine timer(return_time, initial_time)
      real, intent(in)  :: initial_time
      real,  intent(out) :: return_time
      integer finish,rate
      CALL system_clock (COUNT=finish,COUNT_RATE=rate)
      return_time = FLOAT(finish) / FLOAT(rate) - initial_time
      RETURN
      END
