/******************************************************************************
*                                                                             *
*   Update of Ising spin model using simulated tempering,                     *
*   E. Marinari and G. Parisi, Europhys. Lett. 19, 451 (1992).                *
*                                                                             *
*   Initial code by Paul Coddington and Thomas Rose, Summer 1992.             *
*   Dynamic generation of betas and g's by Paul Coddington, Winter 1992/3.    *
*                                                                             *
******************************************************************************/


/******************************************************************************
** TO DO :

** Check_Delta_G():
 * Add changing of delta g's and addition of betas
   (Should we be changing g's or delta_g's???).
 * Check for highest and lowest beta value.
 * Throw away lower values if Num_Betas > MAXBETAS ?

** Change to decent random number generator.
** Make table for finding pbc's.

** Add errors to energies and acceptances to help with checking delta_g's.
** Add possibility of reading in betas and delta_g's and not changing
   them during the run.

******************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>


/****************************  TUNABLE PARAMETERS  ****************************/

int NSweeps		= 10;     /* Number of sweeps per measurement
				   *   (below we will take one "iteration" 
				   *    to mean NSweeps sweeps) */
int NChange		= 1;      /* Number of iterations between attempting
				   *   to update the beta value */
int NZero		= 1;      /* Number of iterations between attempting
				   *   to move away from zero temperature */
int NAnneal		= 1000;   /* Number of initial annealing iterations
				   *   per beta value */
int NCheck		= 1000;   /* Number of iterations between checking 
				   *   acceptances and deciding whether to 
				   *   update the delta_g's and the number 
				   *   of beta values */
int NFail		= 100;    /* Number of unsuccessful attemps to change
				   *   the beta value before we give up and 
				   *   decide the acceptance is too small and
				   *   we need to add another beta value */
int NIters		= 100;    /* Number of iterations (measurements).
				   *   NIters * NSweeps = The total number of 
				   *   sweeps */

int Max_Betas		= 100;    /* Maximum number of beta values */
int Num_Betas		= 2;      /* Number of beta values */
double Min_Beta		= 1.0;    /* Minimum beta value */
double Max_Beta		= 2.0;    /* Maximum beta value */

float Min_Accept 	= 0.3;    /* A new beta value can be added if the 
				   *   acceptance for changing beta is less
				   *   than Min_Accept */
float Zero_Beta_Accept 	= 0.9;    /* beta is considered effectively 0 if the
				   *   acceptance for increasing the energy
				   *   is greater than Zero_Beta_Accept */

/******************************************************************************/


#define TRUE  1
#define FALSE 0

#define MAXROWS 128               /* Maximum lattice size */
#define MAXCOLS 128
int Rows, Cols, Num_Spins;        /* Lattice size */
double Volume;

int Spin[MAXROWS][MAXCOLS];       /* spins */
int MinSpin[MAXROWS][MAXCOLS];    /* minimum energy configuration */
int J[MAXROWS][MAXCOLS][2];       /* couplings */

#define MAXBETAS  1000            /* Maximum number of beta values */
#define MAXENERGY 1000000000      /* Bigger than any possible energy value */

int Beta_Trials[MAXBETAS][2];     /* Number of attempts at changing 
				       to a neighboring beta */
int Beta_Accept[MAXBETAS][2];     /* Metropolis acceptance for changing 
				       to a neighboring beta */
				  /*   0 = neighbor with lower  beta
				   *   1 = neighbor with higher beta */

int    Beta_Calcs[MAXBETAS];      /* Number of measurements at this beta */
int    Beta_Pointer[MAXBETAS];    /* Pointers to betas in ascending order */
double Beta_Value[MAXBETAS];      /* Beta values */
double Delta_G[MAXBETAS];         /* Values of delta g */
double Total_Energy[MAXBETAS];    /* Energy at each beta value */
double Total_Magnet[MAXBETAS];    /* Magnetization at each beta value */
double Total_Accept[MAXBETAS];    /* Acceptance at each beta value */

double Energy, Magnet, Accept; 
double Tot_Energy_Sq, Tot_Magnet_Sq, Tot_Accept_Sq;   
double Error_Energy,  Error_Magnet,  Error_Accept;   

double Last_Energy;               /* Energy value of previous iteration */
double Min_Energy;                /* Minimum energy */

int Energy_Increase;              /* Check whether the energy has increased
				   *   at Max_Beta, which would imply that 
				   *   Max_Beta needs to be increased */

double ranf();                    /* Random number generator */

FILE *DataFile;   /* Any data which might be useful */
FILE *EnerFile;   /* Running list of energies and beta values */
FILE *MiniFile;   /* Running list of minimum energy values */
FILE *BetaFile;   /* Beta values and Delta-Gs */

void Input_Parameters ();
void New_Parameters ();
void Initialize_Random_Numbers (); 
void Initialize_Lattice (); 
void Initialize_Arrays ();
void Order_Betas ();
void Add_New_Beta ();
void Setup_Delta_G ();
int  Check_Delta_G ();
void Metropolis_Update ();
void Beta_Update ();
void Measure ();
void Calculate_Averages ();
void Print_Results ();
void Print_Averages ();
void Print_Betas ();
void Read_Betas ();
void Print_Config ();
void Read_Config ();
void New_Minimum ();


/*******************************************************************************
                                
                                MAIN PROGRAM

*******************************************************************************/

void main()
{

   int Sweep;
   int Iteration; 
   int N_Beta; 
   int Thermalized;     /* Flag for whether beta update is "thermalized" */
   int Beta_Pos;        /* Ordered pointer to current beta value */
   int Beta_Ptr;        /* Pointer to current beta value */
   int Model;           /* Type of spin model (spin glass, etc) */
   int Seed;            /* Seed for random number generator */
   double Beta;         /* Inverse temperature */


   BetaFile = fopen ( "Betas",   "w+");
   DataFile = fopen ( "Data",    "w+");
   EnerFile = fopen ( "Energy",  "w+");
   MiniFile = fopen ( "Minimum", "w+");

   /* Do some initialization */
   Min_Energy = MAXENERGY;
   Input_Parameters ( &Num_Spins, &Rows, &Cols, &Model, &Seed );
   New_Parameters ( &NSweeps, &NChange, &NZero, &NFail, &NAnneal, 
		    &NCheck, &NIters, &Num_Betas );
   Initialize_Random_Numbers ( Seed );
   Initialize_Lattice ( Model );
   Measure ( 0 );
   New_Minimum ( 0 );
   fprintf ( DataFile, "Mag = %f    Energy = %f\n\n", Magnet, Energy);
   fflush  ( DataFile );

   /* 
    * Initial phase - do annealing to calculate initial g values.
    * -----------------------------------------------------------
    */
   fprintf ( DataFile, "ANNEALING PHASE :\n\n" );
   fprintf ( DataFile, "  Beta    Energy\n\n" );
   fflush  ( DataFile );
   Order_Betas ();
   Initialize_Arrays ();
   for ( N_Beta = 0; N_Beta < Num_Betas; N_Beta++ )
   {
      Beta_Ptr = Beta_Pointer[N_Beta];
      Beta = Beta_Value[Beta_Ptr];
      for ( Iteration = 0; Iteration < NAnneal*NSweeps; Iteration++ )
      {
            Metropolis_Update ( Beta );
      }
      for ( Iteration = 0; Iteration < NAnneal; Iteration++ )
      {
         for ( Sweep = 0; Sweep < NSweeps; Sweep++ )
         {
            Metropolis_Update ( Beta );
         }
         Measure ( Beta_Ptr );
	 Print_Results ( Beta );
      }
   }
   fflush ( DataFile );
   Calculate_Averages ();
   Print_Averages ();
   Setup_Delta_G ();
   fflush ( DataFile );

   /* 
    * Tempering phase - dynamically set up g's and number of betas. 
    * -------------------------------------------------------------
    */
   fprintf ( DataFile, "\n\n" );
   fprintf ( DataFile, "TEMPERING PHASE :\n\n" );
   fprintf ( DataFile, "  Beta    Energy  \n\n" );
   fflush  ( DataFile );

   Initialize_Arrays ();
   /* Start with smallest beta value (highest temperature) */
   Beta_Pos = 0;
   Beta_Ptr = Beta_Pointer[Beta_Pos];
   Beta = Beta_Value[Beta_Ptr];
   for ( Iteration = 0; Iteration < NIters; Iteration++ )
   {
      /* 
       * Metropolis update of the configuration.
       */
      for ( Sweep = 0; Sweep < NSweeps; Sweep++ )
      {
         Metropolis_Update ( Beta );
      }
      /* 
       * Do a measurement.
       */
      Measure ( Beta_Ptr );
      Beta_Calcs[Beta_Ptr] ++;
      Print_Results ( Beta );
      if ( (Iteration % NChange) == 0 )
      {
         /* 
          * Update the Beta value.
          */
         Beta_Update ( &Beta_Pos, Last_Energy );
         Beta_Ptr = Beta_Pointer[Beta_Pos];
         Beta = Beta_Value[Beta_Ptr];
      }
      if ( (Iteration % NCheck) == 0 )
      {
         /* 
          * Check if we need to adjust Betas and/or Delta_Gs.
          */
         Check_Delta_G ();
      }
   }
   Calculate_Averages ();
   Print_Averages ();
   Print_Betas ( BetaFile );

   fclose ( BetaFile );
   fclose ( DataFile );
   fclose ( EnerFile );
   fclose ( MiniFile );

   exit(0);
}       



/*******************************************************************************

                            PROCEDURE INPUT_PARAMETERS

   Input required parameters for the particular minimization problem,
   and a seed for the random number generator.

*******************************************************************************/

void 
Input_Parameters( Num_Spins, Rows, Cols, Model, Seed )
int *Num_Spins, *Rows, *Cols, *Model;
int *Seed;
{
   int i; 
   char Sentinel;

   do   
   {
      Sentinel = 0;
      printf ( "Enter the vertical dimensions of the lattice (0 < Y <= %d): ",
	       MAXROWS );
      scanf ( "%d", Rows );
      if ( ( *Rows > MAXROWS ) || ( *Rows <= 0 ) )
      { 
         printf ( "\n\nInput out of range.\n\n" );
         Sentinel = 1;
      }
   } 
   while (Sentinel);

   do
   {
      Sentinel = 0;
      printf("\n\n");
      printf ( "Enter the horizontal dimensions of the lattice (0 < X <= %d): ",
	       MAXCOLS );
      scanf ( "%d", Cols );
      if ( ( *Cols > MAXCOLS ) || ( *Cols <= 0 ) )
      {
            printf ( "\n\nInput out of range.\n\n" );
            Sentinel = 1;
      }
   }
   while (Sentinel);

   *Num_Spins  = (*Rows) * (*Cols);
   Volume = (double)*Num_Spins;
   printf ( "\n\nLATTICE SIZE :    %d X %d = %d\n", *Rows, *Cols, *Num_Spins );
   
   do 
   {
      Sentinel = 0;
      printf ( "\n\nWhich model do you want to simulate?\n" );
      printf ( "1 -- Ferromagnet\n" );
      printf ( "2 -- Spin Glass\n" );
      printf ( "Choice : " );
      scanf ( "%d", Model );
      if ( (*Model != 1) && (*Model != 2) )
      {
         printf ( "\n\nInvalid input\n" );
         Sentinel = 1;
      }
   }
   while (Sentinel);
   fflush(stdout);

   printf ( "\n\nInput a seed for the random number generator:\n\n" );
   scanf ( "%d", Seed );
   
   /* Print input parameters to data file */
   fprintf ( DataFile, "Lattice size :  %d X %d = %d\n", 
		       Rows, Cols, Num_Spins );
   fprintf ( DataFile, "Random number seed = %d\n\n", *Seed );

}
           


/*******************************************************************************

                            PROCEDURE NEW_PARAMETERS

   Input new values to override default values of various parameters.

*******************************************************************************/

void
New_Parameters ( NSweeps, NChange, NZero, NFail, NAnneal, NCheck, NIters, 
		 Num_Betas )
int *NSweeps, *NChange, *NZero, *NFail, *NAnneal, *NCheck, *NIters;
int *Num_Betas;
{
int i; 
int TotSwps;
char Sentinel;

   do 
   {
      printf ( "\n\nEnter the number of sweeps per iteration\n" );
      printf ( "(a measurement is done every iteration) : ");
      scanf  ( "%d", NSweeps );
      printf ( "\n\nEnter the number of iterations between changes : " );
      scanf  ( "%d", NChange );
      printf ( "\n\nEnter the number of iterations for a T=0 change : " );
      scanf  ( "%d", NZero );
      printf ( "\n\nEnter the number of non-changes before failure : " );
      scanf  ( "%d", NFail );
      printf ( "\n\nEnter the number of thermalization iterations : " );
      scanf  ( "%d", NAnneal );
      printf ( "\n\nEnter the number of iterations between checks : " );
      scanf  ( "%d", NCheck );
      printf ( "\n\nEnter the total number of iterations : " );
      scanf  ( "%d", NIters );
      printf ( "\n\n" );
      printf ( "Number of sweeps per iteration       = %d\n", *NSweeps );
      printf ( "Number of iterations between changes = %d\n", *NChange );
      printf ( "Number of iterations for a 0T change = %d\n", *NZero );
      printf ( "Number of non-changes before failure = %d\n", *NFail );
      printf ( "Number of thermalization iterations  = %d\n", *NAnneal );
      printf ( "Number of iterations between checks  = %d\n", *NCheck );
      printf ( "Total number of iterations           = %d\n", *NIters );
      printf ( "\nAre all values correct (y/n) ?" );
      scanf ( "%1s",&Sentinel );
      printf ( "\n\n" );

   } while ( ( Sentinel != 'y' ) && ( Sentinel != 'Y' ) );
   fflush(stdout);
  
   Beta_Value[0] = Min_Beta;
   Beta_Value[1] = Max_Beta;
   printf ( "Restart using old Beta values (y/n) ? " );
   scanf ( "%1s",&Sentinel );
   if ( ( Sentinel == 'y' ) || ( Sentinel == 'Y' ) )
   {
      Read_Betas ();
   }
   else {
      printf ( "Do you want to enter specific Beta values (y/n) ? " );
      scanf ( "%1s",&Sentinel );
      if ( ( Sentinel == 'y' ) || ( Sentinel == 'Y' ) )
      {
         printf ( "Enter the Beta values in increasing order.\n" );
         printf ( "Enter -1 to end input. ( # values <= %d )\n\n", MAXBETAS );
         *Num_Betas = 0;
         i = 0;
         do
         {
            scanf ( "%lf", Beta_Value + i );
            (*Num_Betas)++;
            i++;
         } 
         while ( *( Beta_Value + i - 1 ) > 0.0 );
         (*Num_Betas)--;
         Min_Beta = Beta_Value[0];
         Max_Beta = Beta_Value[*Num_Betas-1];
      }
   }
     
   fprintf ( DataFile, "INITIAL BETA VALUES\n" );
   fprintf ( DataFile, "-------------------\n" );
   for ( i = 0; i < *Num_Betas; i++ ) 
   {
      fprintf ( DataFile, "%14.8lf\n", *( Beta_Value + i ) );
   }
   fflush( DataFile );

   TotSwps = *NIters * *NSweeps;
   fprintf( DataFile, "Number of sweeps per iteration       = %d\n", *NSweeps );
   fprintf( DataFile, "Number of iterations between changes = %d\n", *NChange );
   fprintf( DataFile, "Number of iterations for a 0T change = %d\n", *NZero );
   fprintf( DataFile, "Number of iterations before failure  = %d\n", *NFail );
   fprintf( DataFile, "Number of thermalization iterations  = %d\n", *NAnneal );
   fprintf( DataFile, "Number of iterations between checks  = %d\n", *NCheck );
   fprintf( DataFile, "Total number of iterations           = %d\n", *NIters );
   fprintf( DataFile, "Total number of sweeps               = %d\n", TotSwps );
   fprintf( DataFile, "\n" );
   fprintf( DataFile, "Number of initial beta values  = %d\n", *Num_Betas );

}



/*******************************************************************************

                           PROCEDURE INITIALIZE_LATTICE

*******************************************************************************/

void
Initialize_Lattice( Model )
int Model;
{
   int i, j, dim;

   /* Initialize spins to random values */
   for ( i = 0; i < Rows; i++ ) {
   for ( j = 0; j < Cols; j++ ) {
         if ( ranf() < 0.5 )
            Spin[i][j] =  1;
         else 
            Spin[i][j] = -1;
   }}

   /* Initialize couplings */
   switch (Model)
   {

      case 1  :  /* ferromagnet */
           for ( i = 0; i < Rows; i++ ) {
           for ( j = 0; j < Cols; j++ ) {
              for ( dim = 0; dim < 2; dim++ )           
                 J[i][j][dim] = 1;
	   }}
	   printf("Ferromagnet\n");
           break;

      case 2  :  /* spin glass */
           for ( i = 0; i < Rows; i++ ) {
           for ( j = 0; j < Cols; j++ ) {         
              for ( dim = 0; dim < 2; dim++ )           
                 if ( ranf() < 0.5 )
                    J[i][j][dim] =  1;
                 else 
                    J[i][j][dim] = -1;
	   }}
	   printf("Spin Glass\n");
           break;
 
      default  :
           printf("initialization of couplings is screwed up!\n");
	   fflush(stdout);
	   exit(-3);

   }
}



/*******************************************************************************

                           PROCEDURE INITIALIZE_ARRAYS

*******************************************************************************/

void
Initialize_Arrays ()
{
   int Beta_Ptr, dir;

   for ( Beta_Ptr = 0; Beta_Ptr < MAXBETAS; Beta_Ptr++ )
   {
      Total_Energy[Beta_Ptr]  = 0.0;  
      Total_Magnet[Beta_Ptr]  = 0.0; 
      Total_Accept[Beta_Ptr]  = 0.0;
      Beta_Calcs[Beta_Ptr]  = 0;
      for ( dir = 0; dir < 2; dir++ )
      {
         Beta_Trials[Beta_Ptr][dir] = 0;
         Beta_Accept[Beta_Ptr][dir] = 0;
      }
   }
}



/*******************************************************************************

                           PROCEDURE ORDER_BETAS

   Order the initial beta values in ascending order.

*******************************************************************************/

void 
Order_Betas ( )
{
   int This_Pos, Next_Pos, Beta_Pos;
   int This_Ptr, Next_Ptr;
   double This_Val, Next_Val;

      /*
       * Set up an arbitrary initial ordering.
       */
      for ( Beta_Pos = 0; Beta_Pos < Num_Betas; Beta_Pos++ ) 
      {
            Beta_Pointer[Beta_Pos] = Beta_Pos;
      }
      /* 
       * Now order the pointers doing a simple search ordering.
       */
      for ( Beta_Pos = 0; Beta_Pos < Num_Betas; Beta_Pos++ ) 
      {
         This_Ptr = Beta_Pointer[Beta_Pos];
         This_Val = Beta_Value[This_Ptr];
	 /* 
	  * Search the remaining values for the smallest beta.
	  */
         for ( Next_Pos = Beta_Pos+1; Next_Pos < Num_Betas; Next_Pos++ )
	 {
            Next_Ptr = Beta_Pointer[Next_Pos];
            Next_Val = Beta_Value[Next_Ptr];
            if ( Next_Val < This_Val ) 
	    {
	       This_Val = Next_Val;
	       This_Ptr = Next_Ptr;
	       This_Pos = Next_Pos;
	    }
         }
	 /* 
	  * Swap the current beta pointer with the pointer to
	  * the smallest beta value remaining.
	  */
         Beta_Pointer[This_Pos] = Beta_Pointer[Beta_Pos];
         Beta_Pointer[Beta_Pos] = This_Ptr;
      }
   
      fprintf ( DataFile, "Min beta = %f    Max beta = %f\n", 
		Beta_Value[Beta_Pointer[0]], 
		Beta_Value[Beta_Pointer[Num_Betas-1]] );
} 



/*******************************************************************************

                           PROCEDURE ADD_NEW_BETA

   Add a new beta value to the list. Pass its position (New_Pos) and its 
   value (New_Val). Its pointer is just Num_Betas, which needs to be
   incremented. To put the new beta value in its correct place (so that 
   the values are still in ascending order) we just move all the pointer
   values in the Beta_Pointer array up one, and add the new pointer in the
   new position.
   
   Check that the number of betas does not exceed MAXBETAS, and print an
   error message if it does.
   Maybe here we could throw away a higher beta value??.

*******************************************************************************/

void 
Add_New_Beta ( New_Pos )
int New_Pos;
{
   int New_Ptr, This_Ptr, Next_Ptr;
   double New_Beta, This_Beta, Next_Beta;
   int Beta_Pos;

   /* Increment the number of beta values */
   Num_Betas ++;

   if ( Num_Betas <= MAXBETAS ) 
   {
      /* 
       * Construct new beta value using bisection.
       */
      This_Ptr = Beta_Pointer[New_Pos];
      Next_Ptr = Beta_Pointer[New_Pos+1];
      This_Beta = Beta_Value[This_Ptr];
      Next_Beta = Beta_Value[Next_Ptr];
      New_Beta = 0.5 * (This_Beta + Next_Beta);
      /* 
       * Add the new value to the array of beta values.
       */
      New_Ptr = Num_Betas - 1;
      Beta_Value[New_Ptr] = New_Beta;
      /* 
       * Get the ordering right - put the pointer in the new position. 
       */
      /* First move up all the pointers above the new position  */
      for ( Beta_Pos = Num_Betas; Beta_Pos > New_Pos; Beta_Pos-- )
      {
         Beta_Pointer[Beta_Pos] = Beta_Pointer[Beta_Pos-1]; 
      }
      /* Now add the new pointer at the new position */
      Beta_Pointer[New_Pos] = New_Ptr; 
   }
   else
   {
      printf("ERROR: Num_Betas bigger than MAXBETAS!!\n");
      printf("Failed to add new beta value %f at position %d\n", 
	      New_Beta, New_Pos);
      Num_Betas --;
      /* Maybe here we could throw away a higher beta value?? */
   }
}



/*******************************************************************************

                            PROCEDURE METROPOLIS_UPDATE

*******************************************************************************/

void
Metropolis_Update ( Beta )
double Beta;
{
   int Update;
   int Old_Spin, New_Spin, Spin_Sum;
   double Old_Energy, New_Energy, Energy_Diff;
   int Sum_Accept;
   int i, j;
   int ipos, jpos, ineg, jneg;
   
   Sum_Accept = 0;
   Energy_Increase = FALSE;

   for ( i = 0; i < Rows; i++ )
   {
      if (i < Rows-1) 
         ipos = i+1;
      else 
         ipos = 0;
      if (i > 0)
         ineg = i-1;
      else 
         ineg = Rows-1;
      for ( j = 0; j < Cols; j++ ) 
      {
         if (j < Cols-1)
            jpos = j+1;
         else 
            jpos = 0;
         if (j > 0) 
            jneg = j-1;
         else 
            jneg = Cols-1;

         Old_Spin = Spin[i][j];
         New_Spin = -Old_Spin;

         Spin_Sum = J[i][j][0]    * Spin[ipos][j] + 
		    J[i][j][1]    * Spin[i][jpos] + 
		    J[ineg][j][0] * Spin[ineg][j] + 
		    J[i][jneg][1] * Spin[i][jneg] ;

         Old_Energy = Old_Spin * Spin_Sum;
         New_Energy = New_Spin * Spin_Sum;
         Energy_Diff = -Beta * (New_Energy - Old_Energy);

	 Update = TRUE;
         if ( Energy_Diff > 0.0 ) 
	 {
	    Energy_Increase = TRUE;
	    if ( exp(-Energy_Diff) > ranf() )
	    {
	       Update = FALSE;
	    }
	 }
	 if (Update)
         { 
            Spin[i][j] = New_Spin;
            Sum_Accept ++;
         }
      } /* end of loop over columns */
   } /* end of loop over rows */

   Accept = (double)Sum_Accept / Volume;

}



/*******************************************************************************

			   PROCEDURE BETA_UPDATE

*******************************************************************************/

void
Beta_Update ( Beta_Pos, Tot_Energy )
int *Beta_Pos;
double Tot_Energy;
{
   int Beta_Ptr, New_Beta_Ptr, Old_Beta_Ptr;
   int New_Beta_Pos, Old_Beta_Pos, dir;
   double Boltz_Diff, Beta_Diff, G_Diff;

   Old_Beta_Pos = *Beta_Pos;
   Old_Beta_Ptr = Beta_Pointer[Old_Beta_Pos];
   
   /* Choose new neighboring beta value */
   if ( ranf() < 0.5 )  /* Decrease the beta value */
   {
      dir = 0;
      if ( Old_Beta_Pos != 0 ) 
      {
	 New_Beta_Pos = Old_Beta_Pos - 1;
      }
      New_Beta_Ptr = Beta_Pointer[New_Beta_Pos];
      G_Diff = Delta_G[Old_Beta_Ptr];
   }
   else                   /* Increase the beta value */
   {
      dir = 1;
      if ( Old_Beta_Pos != (Num_Betas - 1) ) 
      {
	 New_Beta_Pos = Old_Beta_Pos + 1;
      }
      New_Beta_Ptr = Beta_Pointer[New_Beta_Pos];
      G_Diff = - Delta_G[New_Beta_Ptr];
   }
   Beta_Trials[Old_Beta_Ptr][dir] ++;

   Beta_Diff = Beta_Value[New_Beta_Ptr] - Beta_Value[Old_Beta_Ptr];
   Boltz_Diff = (Tot_Energy * Beta_Diff) - G_Diff;

#ifdef DEBUG
printf("B = %f  E = %f  E_dB = %f   G_Diff = %f\n",
	Beta_Value[Old_Beta_Ptr], Tot_Energy, Tot_Energy * Beta_Diff, G_Diff );
printf("Boltz_Diff = %f\n", Boltz_Diff);
fflush(stdout);
#endif DEBUG

   if ( ( Boltz_Diff <= 0.0 ) || ( exp ( -Boltz_Diff) > ranf() ) ) {
            *Beta_Pos = New_Beta_Pos;
	    Beta_Accept[Old_Beta_Ptr][dir] ++;
   }

}

   

/*******************************************************************************

                            PROCEDURE SETUP_DELTA_G

   Set up the initial delta_g values, which are taken to be the mid-point
   energies between the two beta values multiplied by delta_beta.

*******************************************************************************/

void
Setup_Delta_G ()
{
   int N_Beta;
   int This_Beta_Ptr, Next_Beta_Ptr;
   double This_Beta, Next_Beta, Delta_Beta, E_Mid_Pt;

   for ( N_Beta = 0; N_Beta < Num_Betas-1; N_Beta++ )
   {
      This_Beta_Ptr = Beta_Pointer[N_Beta];
      This_Beta = Beta_Value[This_Beta_Ptr];
      Next_Beta_Ptr = Beta_Pointer[N_Beta+1];
      Next_Beta = Beta_Value[Next_Beta_Ptr];
      Delta_Beta = This_Beta - Next_Beta;
      E_Mid_Pt = 0.5 * ( Total_Energy[This_Beta_Ptr] + 
                         Total_Energy[Next_Beta_Ptr] );
      Delta_G[This_Beta_Ptr] = E_Mid_Pt * Delta_Beta;
   }
}
 


/*******************************************************************************

                            PROCEDURE CHECK_DELTA_G

   Want to adjust delta_g so that acceptance is around 0.5.
   If this is not possible, then we have to add an extra beta value.

*******************************************************************************/

int  
Check_Delta_G ( )
{
   int Thermalized;
   int Beta_Pos, Beta_Ptr, dir;
   double Beta, Acceptance;

   Thermalized = TRUE;
   for ( Beta_Pos = 0; Beta_Pos < Num_Betas; Beta_Pos++ ) 
   {
      Beta_Ptr = Beta_Pointer[Beta_Pos];
      Beta = Beta_Value[Beta_Ptr];
      for (dir = 0; dir < 2; dir++ )
      {
         if ( Beta_Trials[Beta_Ptr][dir] > 0 )
         {
            /* Is the acceptance bad enough so that we 
             * need to add another beta value?
             */
            /*
	    Acceptance = ;
            */
            if ( Acceptance < Min_Accept ) 
            {
               Thermalized = FALSE;
               Add_New_Beta ( Beta_Pos );
            }
            else 
            {
	       if ( Acceptance > (1-Min_Accept) ) 
	       {
                  Thermalized = FALSE;
                  Add_New_Beta ( Beta_Pos );
	       }
            }
            /* If not, is the acceptance bad enough so that 
             * we should change the value of Delta_G?
             */
         }
      } /* end of loop over directions */
   } /* end of loop over betas */
   return Thermalized;
}


/*******************************************************************************

                            PROCEDURE MEASURE

*******************************************************************************/

void
Measure ( Beta_Ptr )
int Beta_Ptr;
{
   int Sum_Spins, Sum_Magnet, Sum_Energy;
   int i, j;
   int ipos, jpos;
   double Magnet, Energy;

   Sum_Magnet = 0;
   Sum_Energy = 0;

   for ( i = 0; i < Rows; i++ )
   {
      if (i < Rows-1) 
         ipos = i+1;
      else 
         ipos = 0;
 
      for ( j = 0; j < Cols; j++ ) 
      {
         if (j < Cols-1)
            jpos = j+1;
         else 
            jpos = 0;
               
         Sum_Spins = Spin[i][j] * ( J[i][j][0] * Spin[ipos][j] + 
				    J[i][j][1] * Spin[i][jpos] );

         Sum_Magnet += Spin[i][j];
         Sum_Energy -= Sum_Spins;
      }
   }

   Last_Energy = (double)Sum_Energy;

   Energy = (double)Sum_Energy / Volume;
   Magnet = (double)Sum_Magnet / Volume;

   if (Energy < Min_Energy) {
      New_Minimum ( Beta_Ptr );
   }

   Total_Energy[Beta_Ptr] += Energy;
   Total_Magnet[Beta_Ptr] += Magnet;
   Total_Accept[Beta_Ptr] += Accept;

}



/*******************************************************************************

                            PROCEDURE CALCULATE_AVERAGES

*******************************************************************************/

void
Calculate_Averages( )
{
   int Beta_Ptr;
   double Num_Measure;

   for ( Beta_Ptr = 0; Beta_Ptr < Num_Betas; Beta_Ptr++ )
   {
      Num_Measure = (double) Beta_Calcs[Beta_Ptr];
      Total_Energy[Beta_Ptr] /= Num_Measure;
      Total_Magnet[Beta_Ptr] /= Num_Measure;
      Total_Accept[Beta_Ptr] /= Num_Measure;

/*   
      Tot_Energy_Sq[Beta_Ptr] /= Num_Measure;
      Tot_Magnet_Sq[Beta_Ptr] /= Num_Measure;
      Tot_Accept_Sq[Beta_Ptr] /= Num_Measure;

      Error_Energy[Beta_Ptr] = Tot_Energy_Sq[Beta_Ptr] - 
			      (Total_Energy[Beta_Ptr] * Total_Energy[Beta_Ptr]);
      Error_Magnet[Beta_Ptr] = Tot_Magnet_Sq[Beta_Ptr] - 
			      (Total_Magnet[Beta_Ptr] * Total_Magnet[Beta_Ptr]);
      Error_Accept[Beta_Ptr] = Tot_Accept_Sq[Beta_Ptr] - 
			      (Total_Accept[Beta_Ptr] * Total_Accept[Beta_Ptr]);
      Error_Energy[Beta_Ptr] = sqrt ( Error_Energy[Beta_Ptr] / Num_Measure );
      Error_Magnet[Beta_Ptr] = sqrt ( Error_Magnet[Beta_Ptr] / Num_Measure );
      Error_Accept[Beta_Ptr] = sqrt ( Error_Accept[Beta_Ptr] / Num_Measure );
*/   
   }
}
    


/*******************************************************************************

                            PROCEDURE PRINT_RESULTS

*******************************************************************************/

void
Print_Results ( Beta )
{
   fprintf ( DataFile, "%f  %10.8lf\n", Beta, Energy );
}
 


/*******************************************************************************

                            PROCEDURE PRINT_AVERAGES

*******************************************************************************/

void
Print_Averages ( )
{
   int N_Beta, Beta_Ptr;

   fprintf ( DataFile, "\n" ); 
   fprintf ( DataFile, "  BETA     ENERGY    MAGNETIZATION   ACCEPTANCE \n" ); 
   for ( N_Beta = 0; N_Beta < Num_Betas; N_Beta++ )
   {
      Beta_Ptr = Beta_Pointer[N_Beta];
      fprintf ( DataFile, "\n" );
      if ( Beta_Calcs[Beta_Ptr] > 0 ) 
      {
         fprintf ( DataFile, "%lf   ",     Beta_Value[Beta_Ptr] ); 
         fprintf ( DataFile, "%10.8lf   ", Total_Energy[Beta_Ptr] ); 
         fprintf ( DataFile, "%10.8lf   ", Total_Magnet[Beta_Ptr] ); 
         fprintf ( DataFile, "%10.8lf   ", Total_Accept[Beta_Ptr] ); 
      }
   }
   fprintf ( DataFile, "\n\n" ); 
   fprintf ( DataFile, "MINIMUM ENERGY = %10.8lf\n", Min_Energy ); 
}



/*******************************************************************************

                            PROCEDURE PRINT_BETAS

*******************************************************************************/

void
Print_Betas ( FileName )
FILE *FileName;
{
   int N_Beta, Beta_Ptr, dir;
   double Acceptance;

   fprintf ( FileName, "  BETA     DELTA_G    NEG ACCEPT   POS ACCEPT \n\n" ); 
   for ( N_Beta = 0; N_Beta < Num_Betas; N_Beta++ )
   {
      Beta_Ptr = Beta_Pointer[N_Beta];
      fprintf ( FileName, "%lf   ",     Beta_Value[Beta_Ptr] ); 
      fprintf ( FileName, "%10.8lf   ", Delta_G[Beta_Ptr] ); 
      for ( dir=0; dir<2; dir++ ) 
      {
         if ( Beta_Trials[Beta_Ptr][dir] > 0 ) 
         {
            Acceptance = (double)Beta_Accept[Beta_Ptr][dir] / 
	   	         (double)Beta_Trials[Beta_Ptr][dir];
         }
         else 
         {
            Acceptance = -1.0;
         }
         fprintf ( FileName, "%10.8lf   ", Acceptance );
      }
      fprintf ( FileName, "\n" ); 
   }
}



/*******************************************************************************

                            PROCEDURE READ_BETAS

   Read beta and delta_g values from the previous run, to save having
   to generate them again.

*******************************************************************************/

void
Read_Betas ()
{
}



/*******************************************************************************

                            PROCEDURE PRINT_CONFIG

   Print out the configuration with the current minimum energy value.

*******************************************************************************/

void
Print_Config ()
{
   int i,j;

   FILE *ConfFile;
   ConfFile = fopen ( "Config", "w+");

   for ( i = 0; i < Rows; i++ )
   {
      for ( j = 0; j < Cols; j++ )
      {
         if ( Spin[i][j] ==  1 ) 
         {
            fprintf ( ConfFile, "1" ); 
         }
         else {
            fprintf ( ConfFile, "0" ); 
         }
      }
      fprintf ( ConfFile, "\n" ); 
   }
   fclose ( ConfFile );
}



/*******************************************************************************

                            PROCEDURE NEW_MINIMUM

   Update all the files and variables with the new minimum values

*******************************************************************************/

void
New_Minimum ( Beta_Ptr )
int Beta_Ptr;
{
   int i,j;
   double Beta;

   Min_Energy = Energy;
   Beta = Beta_Value[Beta_Ptr];
   if ( Beta == Max_Beta )
   {
      /* Print out minimum energy */
      fprintf ( MiniFile, "%10.6f\n", Min_Energy );
      fprintf ( DataFile, "New minimum -- %10.6f\n", Min_Energy );
      printf ( "New minimum -- %10.6f\n", Min_Energy );

      /* Print out minimum energy configuration */
      for ( i = 0; i < Rows; i++ ) {
      for ( j = 0; j < Cols; j++ ) {
         MinSpin[i][i] = Spin[i][j];
      }}
      Print_Config();
   }
}



/*******************************************************************************

	                RANDOM NUMBER GENERATOR

*******************************************************************************/

void
Initialize_Random_Numbers( seed )
int seed;
{
   srand( seed );
}


double ranf()
{
 double temp;

 temp = (double)rand()/(double) RAND_MAX; 
 if (temp == 1.0) temp = 0.0;

 if ( (temp >1.0) || (temp < 0 ) ) {
     perror("ERROR: ranf is bad");
     printf("ranf = %lf \n",temp);
     exit(-2);
 }

 return(temp);

}

