/*+---------------------------------------------------------------------+
  |			File: lattice.c					|
  +---------------------------------------------------------------------+*/

#include "lattice.h" 

#define	MAXrows 256
#define MAXcols 256

static 	int 	Spins[MAXrows][MAXcols];
static	int 	J[MAXrows][MAXcols][2];

static 	int 	metropolis();


void 	init_lattice()
/*+---------------------------------------------------------------------+
  |	Initializes spins and the couplings on a rows x cols lattice.	|
  |	If parameter spin_galss is true J is set to +/- 1 randomly;     |
  |	otherwise it is set to 1.  The model with J=1 has an analytic 	|
  |	solution.							|
  +---------------------------------------------------------------------+*/
{
  	int 	i,j;
	
	assert(pRows()<=MAXrows && pCols()<=MAXcols);
	
	/* 	Arrange the spins randomly 				 */

  	for (i=0; i<pRows(); i++)
    	   for (j=0; j<pCols(); j++) 
      	      Spins[i][j] = (kiss()<0.5) ? 1 : -1; 
      	      
    	/*	Arrange the coupling, J, randomly for simulating 
    		the Ising Spin Glass and equal to 1 for the Ising
    		Model.							 */ 
    		
    	if (!pIsing()) {
    	   for (i=0; i<pRows(); i++)
    	      for (j=0; j<pCols(); j++) {  	  
      	         J[i][j][0] = (kiss()<0.5) ? 1 : -1;
      	         J[i][j][1] = (kiss()<0.5) ? 1 : -1;
      	      }
      	   return;
      	}
      	
    	for (i=0; i<pRows(); i++)
    	   for (j=0; j<pCols(); j++)  	  
      	      J[i][j][0] = J[i][j][1] = 1;
}


void 	thermalize(beta,N)
	double	beta;
	int	N;
{
	int 	i;
	int	accept;
	
	for (i=0; i<N; i++) {
	   accept = metropolis(beta);
     	}
}


static  int	metropolis(beta)
/*+---------------------------------------------------------------------+
  |	Metropolis algorithm for system evolution as described in	|
  |	Metropolis et al, "Equation of State Calculation by Fast 	|
  |	Computing Machines," J. Chem. Phyis., (1953) vol. 21, 		|
  |	pp. 1087-1092.							|
  +---------------------------------------------------------------------+*/
  	double 	beta;
{
	int 	i,j,rows,cols;
	int	old_spin, new_spin, sum;
	int 	ipos, ineg, jpos, jneg;
	int	accept;
	double	oldE, newE, deltaE;
	
 	accept = 0;
 	rows = pRows();
 	cols = pCols();
 	
  	for (i=0; i<rows; i++) {
    	   ipos = (i<rows-1) ? i + 1 : 0;
    	   ineg = (i>0) ? i - 1 : rows - 1;
	   for (j=0; j<cols; j++) {
      	      jpos = (j<cols-1) ? j + 1 : 0;
              jneg = (j>0) ? j - 1 : cols - 1;
      	      old_spin = Spins[i][j];
      	      new_spin = -old_spin;
      	      sum = J[i][j][0]    * Spins[ipos][j] +
                    J[i][j][1]    * Spins[i][jpos] +
                    J[ineg][j][0] * Spins[ineg][j] +
                    J[i][jneg][1] * Spins[i][jneg];
              oldE   = -beta * old_spin * sum;
      	      newE   = -beta * new_spin * sum;
      	      deltaE = newE - oldE;
      		
      	      if ( (deltaE<=0.0) || (exp(-deltaE)>kiss()) ) {
                  Spins[i][j] = new_spin;
                  accept ++;
	      }
	      update_hist(old_spin,Spins[i][j]);
	         
    	   }
    	}
	return accept;
}


struct	Stats	calculate()
/*+---------------------------------------------------------------------+
  |	Calculate the internal energy and the magnetization for the  	|
  |	entire lattice.							|
  +---------------------------------------------------------------------+*/
{
	int	i,j,rows,cols;
	int	ipos, jpos;
	struct	Stats	s;
	double	energy, magnet;
	int	sum;
	
	energy = magnet = 0.0;
	rows = pRows();
	cols = pCols();
	
	for (i=0; i<rows; i++) {
	   ipos = (i<rows-1) ? i + 1 : 0;
	   for (j=0; j<cols; j++) {
      	      jpos = (j<cols-1) ? j + 1 : 0;
      	      sum = Spins[i][j] * 
               (J[i][j][0] * Spins[ipos][j] +
                J[i][j][1] * Spins[i][jpos]);
      	      energy += -1.0 * (double)sum;
      	      magnet += (double)Spins[i][j];
    	   }
  	}
  	s.E = energy;
  	s.Magnet = magnet;
  	return s;
}

