/*+---------------------------------------------------------------------+
  |				File: temper.c				|
  +---------------------------------------------------------------------+*/


#include "temper.h"

static	double	deltaG();
static	void	update_stats();
static	void	disp_stats();
static	void	calc_stats();
static	void	ferd_stats();
static	void	temper();
static	int	tsweeps();
static	void	dinfo();
static	void	init_tmpr();

static	struct	{
	double 	minE;
	double	Z[MAXBETAS];
	double	U[MAXBETAS];
	int	sweeps[MAXBETAS];
	} res;

static	double	dGs[MAXBETAS];
static	int	dGcnt[MAXBETAS];
static	double	Hs[MAXBETAS];
static	int	Hcnt[MAXBETAS];


int	main(argc,argv)
/*+---------------------------------------------------------------------+
  |	Main routine for simulated tempering.  Performs various 	|
  |	initializations, invokes the tempering routine, temper(), and	|
  |	invokes some of the available output routines.			|
  +---------------------------------------------------------------------+*/
	int	argc;
	char	*argv[];
{
  	init_params(argc,argv);
	init_lattice();
	init_betas();
	init_tmpr();
	init_hist();
	init_busy(pNSweeps(),pDebug());
	temper();
	calc_stats();
	if (pDebug()) disp_params();
  	ferd_stats();
  	write_hist();
  	dispB();
}


static  void	temper()
/*+---------------------------------------------------------------------+
  |									|
  |	Performs a Monte Carlo simulation of an Ising spin system as	|
  |	prescribed by Marinari and Parisi, Europhys. Lett., 19:451,1992.|
  |									|
  |	o If the debug switch is turned on various simulation statistics|
  |	  will be printed to the screen at each sweep.			|
  |	o The routine busy() is cosmetic.  It displays a "graphical"	|
  |	  summary of how far the simulation has progressed.		|
  |									|
  +---------------------------------------------------------------------+*/
	
{
  	struct	Stats	s1, s2;	
  	double	dG, 	dHt, dB;
  	double	beta2;
  	int	i, ns; 	
	
	for (i=0; i<pNSweeps(); i++) {

	   ns = tsweeps(curB());
	   thermalize(curB(),ns);
	   s1 = calculate();
	   update_stats(&s1,ns);	   
	   
	   beta2 = tryB();
	   thermalize(beta2,tsweeps(beta2));
	   s2 = calculate();

	   dB = deltaB();
	   dG = deltaG(&s1,&s2);
	   dHt = s1.E*dB - dG;
	   	   
	   if ((dHt<=0.0) || (exp(-dHt)>kiss())) 
	      updateB();
	   (pDebug()) ? dinfo(i,dHt,dG,curB(),beta2) : busy(i);
	}
}
       	

static  double	deltaG(s1,s2)
/*+---------------------------------------------------------------------+
  |	Returns deltaG.  This parameter is calcuated by averaging the   |
  |	average energy at the current beta with the average energy at 	|
  |	the proposed beta.  Note the optional iterative refinement 	|
  |	procedure.							|
  +---------------------------------------------------------------------+*/

	struct	Stats	*s1;
	struct	Stats 	*s2;
{
	double	dG,dB;
	int	i,n;
		
	dB = deltaB();
	dG = dB*(0.5*(s1->E+s2->E));
	if (tryDir()==UP) {
	   i = indexB() - 1;
	   dG = -dG;
	   }
	else
	   i = indexB() + 1;
	assert(i>=0 && i<MAXBETAS);
	if (!pZoom()) return dGs[i] = dG;
	n = dGcnt[i];
	dG = (dGs[i]*n + dG)/(double)(n+1);
	dGcnt[i]++;
	return dGs[i] = dG;
}


static	int	tsweeps(beta)
/*+---------------------------------------------------------------------+
  |	This function is experimental.  It returns the number of sweeps |
  |	which is a function of the temperature.  The motivation is to   |
  |	sample more at lower temperatures.				|
  +---------------------------------------------------------------------+*/
	double	beta;
{
	if (beta<0.1) return pTSweeps();
	return pTSweeps() + (int)pow(pTSweeps(),beta+0.3);
}


static  void	update_stats(s,N)
/*+---------------------------------------------------------------------+
  | 	Updates various statistics; iteratively refines the		|
  |	Hamiltonians.							|
  +---------------------------------------------------------------------+*/
	int	N;
	struct	Stats	*s;
{
	int	i,n;
	
	i = indexB();
	n = Hcnt[i];
	Hs[i] = (Hs[i]*n + s->E)/(double)(n+1);
	Hcnt[i]++;
	res.minE = (res.minE < s->E) ? res.minE : s->E;
	res.U[i] = Hs[i];	
	res.sweeps[i] += N;
}


static  void	calc_stats()
/*+---------------------------------------------------------------------+
  |	Calculates internal energies per spin.                          |
  +---------------------------------------------------------------------+*/
{
	int	i;
	double	sz;
	
	sz = (double)(pRows()*pCols());	
	for (i=0; i<=lastBi(); i++) 
	   res.U[i] /= sz;
}


static  void	init_tmpr()
{
	int	i;
	
	res.minE = 1000.0;
	for (i=0; i<MAXBETAS; i++) {
	   res.U[i] = res.Z[i] = 0.0;
	   res.sweeps[i] = 0;
	   dGs[i] = Hs[i] = 0.0;
	   dGcnt[i] = Hcnt[i] = 0;
	}	
}


static  void	disp_stats()
{
	int	i;
	
	printf("\n\n");
	printf("\t\t Simulation statistics:\n\n");
	printf("\t\t        Minimum Energy --> %f\n\n", res.minE);
}


static  void	ferd_stats()
/*+---------------------------------------------------------------------+
  |	Displays a comparison of statistics accumulated by simulation 	|
  |	and analysis.  Internal energy (per spin) is the measured 	|
  |	quantity.							|
  +---------------------------------------------------------------------+*/
{
	int 	i,ts;
	double	fU;
	
	ts = 0;
	for (i=0;i<lastBi();i++)
	   ts += res.sweeps[i];
	   
	printf("\n\n");
	printf("\t\t Simulation statistics:\n\n");
	printf("\t\t        Minimum Energy --> %f\n", res.minE);
	printf("\t\t          Total sweeps --> %d\n\n",ts);
	
	printf("\t\t Numerical and Analytic results:\n\n");
	printf("  Beta       Int Engy        Ferdinand and Fisher  Err   \t  Sweeps\n");
	printf("  ----       --------        --------------------  ---   \t  ------\n\n");
	for (i=0; i<=lastBi(); i++) 
	   printf("  %10.8f %15.12f %15.12f \t   %10.8f %10d\n", 
	           thisB(i),res.U[i],fU=ferdU(thisB(i),pRows()),
	           fU/res.U[i],res.sweeps[i]);
}


static  void	dinfo(i,dHt,dG,cur,next)
/*+---------------------------------------------------------------------+
  |	Displays simulation information.  cur and next correspond to 	|
  |	the current and the proposed betas, respectively.  Acceptance of|
  |	the proposed beta, next, has occurred if cur equals next.	|
  +---------------------------------------------------------------------+*/
	int	i;
	double	dHt,dG;
	double	cur,next;
{
	if (cur==next)
	   printf("[%4d] yeah: %f dHt = %12.7f dG = %12.7f\n",i,cur,dHt,dG);
	else
	   printf("[%4d] shit: %f dHt = %12.7f dG = %12.7f\n",i,next,dHt,dG);
}
