
/*+---------------------------------------------------------------------+
  |				File: ferd.c				|
  |									|
  |	Analytic solution for finite 2D Ising spin glass with periodic  |
  |	boundary conditions.						|
  |									|
  |	Ferdinand, A. E. and Fisher, M. E. "Bounded and Inhomogeneous   |
  |	Ising Models. I. Specific-Heat Anomaly of a Finite Lattice,"	|
  |	Phys. Rev., vol. 185, no. 2, 1969, (832-846).			|
  |									|
  |	Originally coded by Bernd Berg (berg@scri.fsu.edu)		|
  |	Matthew Shore, Mar. 1994.					|
  |									|
  +---------------------------------------------------------------------+*/
 
#include "ferd.h"

#define	MAX	500
#define	SET	321

static	struct {
	int	nl;		/* Lattice sites (???) 		*/
	double	beta;		/* 1/kT				*/
	double	bk;
	int	ll;
	int	iflag;		/* set when initialized		*/
	} P;
	
static	struct {
	double 	gm[MAX];
	double	gmpr[MAX];
	double	gm2pr[MAX];
	} G;
	
static	struct {
	double	ZT;
	double 	ZloPf;
	double	ZTpr;
	double	ZT2pr;
	} partfn;
	
static	void	init();
static	void	Gamma();
static	void	PartFn();


double	ferdU(beta,size)
/*+---------------------------------------------------------------------+
  |	Given beta and model size (e.g., for 10x10 lattice size==10)	|
  |	returns the corresponding internat energy per spin.		|
  +---------------------------------------------------------------------+*/
	double	beta;
	int	size;
{
	assert(beta>=0.0);
	assert(size>0&&size<MAX);
	P.beta = beta;
	P.nl = size;
	P.iflag = SET;
	init();
	Gamma();
	PartFn();
	return -1.0/tanh(P.bk) - partfn.ZTpr/(partfn.ZT*P.nl*P.nl);
}


double	ferdZ(beta,size)
/*+---------------------------------------------------------------------+
  |	Given beta and model size (e.g., for 10x10 lattice size==10)	|
  |	returns the corresponding partition function. 			|
  +---------------------------------------------------------------------+*/
	double	beta;
	int	size;
{
	assert(beta>=0.0);
	assert(size>0&&size<MAX);
	P.beta = beta;
	P.nl = size;
	P.iflag = SET;
	init();
	Gamma();
	PartFn();
	return partfn.ZT;
}


void	ztest()
{
	float	B;
	int	s;
	
	printf("\n\n");
	printf("\t\t            Enter Beta --> ");
	fscanf(stdin,"%f",&B);
	P.beta = (double)B;
	printf("\t\t    Enter lattice size --> ");
	fscanf(stdin,"%d",&s);
	P.nl = s;
	P.iflag = SET;
	init();
	Gamma();
	PartFn();
}
	

void	dispZ()
{
	printf("\n\n");
	printf("\t\t       Z --> %e\n",partfn.ZT);
	printf("\t\t     Zpr --> %e\n",partfn.ZTpr);
	printf("\t\t    Z2pr --> %e\n",partfn.ZT2pr);
}


static  void	init()
{
	int	i;
	
	assert(P.iflag==SET);
	P.bk = 2.0 * P.beta;
	P.ll = P.nl*2-1;
	partfn.ZT = partfn.ZTpr = partfn.ZT2pr = 0.0;
	for (i=0; i<MAX; i++) 
	   G.gm[i] = G.gmpr[i] = G.gm2pr[i] = 0.0;
}	

static  void	Gamma()
{
	double	tanhsq,taprn;
	double 	c[MAX];
	double	csq,cpr,c2pr,c2pr1;
	int	i;
	double	pi;
	
	assert(P.iflag==SET);
	csq = cpr = c2pr = c2pr1 = 0.0;
	tanhsq = taprn = 0.0;
	
	G.gm[0] = P.bk + log(tanh(P.bk/2.0));
	G.gmpr[0] = 2.0 + 1.0/tanh(P.bk/2.0) - tanh(P.bk/2.0);
	tanhsq = tanh(P.bk/2.0)*tanh(P.bk/2.0);
	G.gm2pr[0] = -tanhsq - (1.0/tanhsq);
	pi = 4.0*atan(1.0);
	
	for (i=1;i<=P.ll;i++) {
	
	   c[i] = cosh(P.bk)*(1.0/tanh(P.bk)) - cos(pi*(double)i/(double)P.nl);
	   csq = pow((c[i]*c[i] - 1.0),0.5);
	   G.gm[i] = log(c[i]+csq);
	   
	   taprn = 1.0 - pow(1.0/tanh(P.bk),2.0);
	   cpr = 2.0*sinh(P.bk)/tanh(P.bk) + 2.0*cosh(P.bk)*taprn;
	   G.gmpr[i] = cpr/csq;
	   
	   c2pr1 = sinh(P.bk) - cosh(P.bk)/tanh(P.bk);
	   c2pr1 *= taprn*2.0;
	   c2pr = c2pr1 + cosh(P.bk)/tanh(P.bk);
	   c2pr *= 4.0;
	   G.gm2pr[i] = (c2pr - G.gmpr[i]*G.gmpr[i]*c[i])/csq;
	   
	}
}


static  void	PartFn()
{
	double	Z1,Z2,Z3,Z4,Z0;
	double 	Zpr1, Zpr2, Zpr3, Zpr4;
	double	Z2pr1, Z2pr2, Z2pr3, Z2pr4;
	double	Z2pr11,Z2pr12,Z2pr13,Z2pr14,Z2pr21,Z2pr31;
	double	Z2pr22,Z2pr23,Z2pr32,Z2pr33,Z2pr24,Z2pr34;
	double	ZT,ZTpr,ZT2pr;
	double	ZloZ0,ZloPf;
	int	i,nz;
	
	nz = P.nl - 1;	
	Z1 = Z2 = Z3 = Z4 = 1.0;	
	Zpr1 = Zpr2 = Z2pr1 =Z2pr21=Z2pr31= 0.0;
	Z2pr2=Z2pr22=Z2pr32=0.0;
	Zpr3=Z2pr3=Z2pr23=Z2pr33=0.0;
	Zpr4=Z2pr4=Z2pr24=Z2pr34=0.0;
	
	Z0 = pow(10.0,-(double)P.nl/4.0);
	ZloZ0 = (double)P.nl*log(Z0);

			/* partial partition function: Z1 */
	
	for (i=0;i<=nz;i++) 
	   Z1 *= 2.0*cosh(G.gm[2*i+1]*P.nl/2.0)*Z0;
	
	for (i=0;i<=nz;i++) 
	   Zpr1 += G.gmpr[2*i+1]*tanh(G.gm[2*i+1]*P.nl/2.0);
	Zpr1 *= Z1*P.nl/2.0;
	
	Z2pr11 = (Zpr1/Z1)*(Zpr1/Z1);
	for (i=0; i<=nz; i++)
	   Z2pr21 += G.gm2pr[2*i+1]*tanh(G.gm[2*i+1]*P.nl/2.0);
	Z2pr21 *= P.nl/2.0;
	
	for (i=0;i<=nz;i++)
	   Z2pr31 += pow(G.gmpr[2*i+1]/cosh(G.gm[2*i+1]*P.nl/2.0),2.0);
	Z2pr31 *= P.nl*P.nl/4.0;
	Z2pr1 = (Z2pr11+Z2pr21+Z2pr31)*Z1;
	
			/* partial partition function: Z2 */
			
	for (i=0;i<=nz;i++)
	   Z2 *= 2.0*sinh(G.gm[2*i+1]*P.nl/2.0)*Z0;
	   
	for (i=0;i<=nz;i++)
	   Zpr2 += G.gmpr[2*i+1]/tanh(G.gm[2*i+1]*P.nl/2.0);
	Zpr2 *= Z2 * P.nl/2.0;
	
	Z2pr12 = (Zpr2/Z2)*(Zpr2/Z2);
	for (i=0;i<=nz;i++) 
	   Z2pr22 += G.gm2pr[2*i+1]/tanh(G.gm[2*i+1]*P.nl/2.0);
	Z2pr22 *= P.nl/2.0;
	
	for (i=0;i<=nz;i++)
	   Z2pr32 += pow(G.gmpr[2*i+1]/sinh(G.gm[2*i+1]*P.nl/2.0),2.0);
	Z2pr32 *= P.nl*P.nl/4.0;
	Z2pr2 = (Z2pr12+Z2pr22-Z2pr32)*Z2;
	
			/* partial partition function: Z3 */
			
	for (i=0;i<=nz;i++)
	   Z3 *= 2.0*cosh(G.gm[2*i]*P.nl/2.0)*Z0;
	   
	for (i=0;i<=nz;i++)
	   Zpr3 += G.gmpr[2*i]*tanh(G.gm[2*i]*P.nl/2.0);
	Zpr3 *= Z3*P.nl/2.0;
	
	Z2pr13 = (Zpr3/Z3)*(Zpr3/Z3);
	for (i=0;i<=nz;i++)
	   Z2pr23 += G.gm2pr[2*i]*tanh(G.gm[2*i]*P.nl/2.0);
	Z2pr23 *= P.nl/2.0;
	
	for (i=0;i<=nz;i++)
	   Z2pr33 += pow(G.gmpr[2*i]/cosh(G.gm[2*i]*P.nl/2.0),2.0);
	Z2pr33 *= P.nl*P.nl/4.0;
	Z2pr3 = (Z2pr13+Z2pr32+Z2pr33)*Z3;
	
			/* partial partition function: Z4 */
			
	for (i=0;i<=nz;i++)
	   Z4 *= 2.0*sinh(G.gm[2*i]*P.nl/2.0)*Z0;
	   
	for (i=0;i<=nz;i++)
	   Zpr4 += G.gmpr[2*i]/tanh(G.gm[2*i]*P.nl/2.0);
	Zpr4 *= Z4*P.nl/2.0;
	
	Z2pr14 = (Zpr4/Z4)*(Zpr4/Z4);
	for (i=0;i<=nz;i++)
	   Z2pr24 += G.gm2pr[2*i]/tanh(G.gm[2*i]*P.nl/2.0);
	Z2pr24 *= P.nl/2.0;
	for (i=0;i<=nz;i++)
	   Z2pr34 += pow(G.gmpr[2*i]/sinh(G.gm[i*2]*P.nl/2.0),2.0);
	Z2pr34 *= P.nl*P.nl/4.0;
	Z2pr4 = (Z2pr14+Z2pr24-Z2pr34)*Z4;
	
			/* canonical partition function and derivatives */
	
	partfn.ZT = Z1 + Z2 + Z3 + Z4;
	partfn.ZTpr = Zpr1 + Zpr2 + Zpr3 + Zpr4;
	partfn.ZT2pr = Z2pr1 + Z2pr2 + Z2pr3 + Z2pr4;
	
	partfn.ZloPf = log(0.5*ZT) + (P.nl*P.nl/2.0)*log(2.0*sinh(P.bk));
	partfn.ZloPf -= ZloZ0;
		  	
}
