/* ucdp.c */

/* This program implements a forward dynamic programming algorithm to 
   explicitly work with the power generation unit commitment problem.  
   In particular, this sequential version of the unit commitment dynamic 
   programming program is designed to implement example 5E on pp. 125-132.  
   This program will combine the heuristic of using a priority list with 
   taking all possible combinations of generating units.  The all possible 
   combinations are for a limited number of generators that are distributed 
   around the number of "must run"  and prioritized generators that 
   would be required to meet the load plus spinning reserve.  "NumCombo" 
   combinations of generators will be considered, with "NumLTLoad" generators 
   in the combinations less than the number required for load plus spinning 
   reserve.  Use of this heuristic will greatly reduce the number of 
   infeasible solutions examined by ensuring that "must run" generators 
   are included and unavailable generators are avoided, while supplying 
   adequate choices to handle such situations as ramped start-up costs. */

/* All input data for this program is placed in the file "ucdpin.dat" 
   and all output data will be written to file "ucdpout.dat". */

/* In this program, descriptive names have been used for variables and 
   function names to simplify reading the code.  The convention used here 
   is from the object oriented programming world, where the first letter 
   in words are capitalized, while words are concatenated without hyphens 
   or underscores. Whenever possible this convention will even be carried 
   to index variables, to improve the "readability" of the code. */

#include <stdio.h>
#include <math.h>
#include <string.h>

/* define some constants */

#define MAXPERIODS    168
#define MAXSTATES     256
#define MAXNUMUNITS   50
#define MAXNUMCOMBOS  8
#define INFINITY      1.0E10
#define YES           1
#define NO            0
#define ONLINE        1
#define OFFLINE       0
#define PREV          0
#define PRES          1

/* Define a matrix of values that depict whether or not particular 
   generators are turned on or turned off.  This value is the binary 
   representation of the value of the first index.  There will be a 
   trade off calculating these values once and storing them in a matrix 
   or calculating them each time they are required.  The values will be 
   required often, so they will be stored.  Because the values are 
   limited to either 0 for off or 1 for on, short integers will be used. */

short BinUnitID[MAXSTATES][MAXNUMCOMBOS+1];


/* Define a structure that represents a state or node in the dynamic 
   programming problem.  A state is completely defined by three items: 

      1. least total cost to arrive at this state (LeastTotalCost),
      2. a character representation of the binary generating unit status 
         (on-off) for this state (GenStatus[]), and
      3. pointer to the state in the previous period that is
         potentially in the optimal schedule (BackPointer).

   The generating unit combination at a state is defined by combining 
   the priority list heuristic with taking all combinations of a fixed 
   number of generators distributed around the combination required for 
   load plus spinning reserve. */

struct State {float LeastTotalCost;
              char  GenStatus[MAXNUMUNITS+2];
              int   BackPointer};


/* In the dynamic programming algorithm, data at each state in each 
   stage must be saved.  So, define a matrix of states for various 
   periods.  The 0th period is the initial stage, while there are 
   "NumPeriods" periods to be examined.  Consequently, there are a 
   total of "NumPeriods+1" periods.  All "2^NumCombo" states will be  
   examined each period. */

struct State States[MAXPERIODS+1][MAXSTATES];


/* A generating unit is uniquely defined by the following structure.  
   Elements of the structure are:

      1. minimum output in (MW) (MinOutput),
      2. maximum output in (MW) (MaxOutput),
      3. incremental heat rate in (Btu/mWh) (IncHeatRate),
      4. no-load cost in (Btu/h) (NoLoadCost),
      5. full-load average cost in ($/mWh) (FullLoadAveCost),
      6. fuel cost in ($/MBtu) FuelCost,
      7. minimum up time in (h) (MinUpTime),
      8. minimum down time in (h) (MinDownTime),
      9. hot start-up costs in ($) (HotStartCost),
     10. cold start-up-costs in (MBtu) (ColdStartCost),
     11. fixed cold start-up-costs in ($) (ColdStartFixedCost),
     12. cold start in (h) (ColdStart),
     13. thermal time constant (ThermTimeConst),
     14. cost of maintaining unit at operating temperature 
         in (Btu/h) (MaintainTempCost),
     15. indicator of whether unit is online or offline (OnOffLine[][]),
     16. time on or off line in (h) (TimeOnOffLine[][]),
     17. indicator of whether unit is must-run (MustRun[][]), 
     18. indicator of whether unit is available (Available[][]), and
     19. unit production in (MW) (UnitProduction[]).

   These unit characteristics are from table 5.4 in "Power Generation, 
   Operation and Control", and they will be modified to suit the 
   requirements of Prof. Y. Wallach.

   Some hysterias on generating unit information must be maintained, but 
   only for the previous state.  Consequently, OnOffLine[][], 
   TimeOnOffLine[][], MustRun[][], and Available[][] are for only 
   kept for the previous state and modified during the present state. */

struct Unit {float MinOutput;
             float MaxOutput;
             float IncHeatRate;
             float NoLoadCost;
             float FullLoadAveCost;
             float FuelCost;
             int   MinUpTime;
             int   MinDownTime;
             float HotStartCost;
             float ColdStartCost;
             float ColdStartFixedCost;
             int   ColdStart;
             float ThermTimeConst;
             float MaintainTempCost;
             int   OnOffLine[2][MAXSTATES];
             int   TimeOnOffLine[2][MAXSTATES];
             int   MustRun[2][MAXSTATES];
             int   Available[2][MAXSTATES];
             float UnitProduction[MAXSTATES]};


/* Declare an array of Units - units are numbered starting at 1 */

struct Unit Units[MAXNUMUNITS+1];


/* Declare a structure that matches the full load average cost with
   the index to that value */

struct OrdUnit {int FullLoadAveCost;
                int Pointer};


/* Declare an array of ordered units */

struct OrdUnit OrdUnits[MAXNUMUNITS+1];


/* Declare an array of possible generating units */

int PossUnits[MAXNUMUNITS+1];

int NumUnits;     /* number of generating units to be examined */
int NumStates;    /* number of states to be examined per period */
int NumPeriods;   /* number of periods to be examined */
int NumCombos;    /* number of generating unit combinations to be 
                     examined.  NumStates should equal 2^NumCombos */
int NumLTLoad;    /* number of generating units to backtrack on unit
                     priority list to when setting up the number of
                     combinations. */


FILE *OutputFile; /* pointer to the output file stream */


/* Declare an array of load patterns for each of NumPeriods */

float LoadPattern[MAXPERIODS+1];


/* define user-supplied functions */
/* function names are designed to be self-explanatory */

void  InitStaticData();
void  InitMatrixOfStates();
void  InitBinUnitID();
void  InitOrderedUnitList();
void  SortOrderedUnits();
void  CalcLeastTotalCost();
void  GeneratePossibleUnitList();
void  BinaryRep();
float ProductionCost();
void  EconomicDispatch();
float TransitionCost();
void  TraceOptimalSchedule();
/* operations in the main routine are abstracted by extensive use of
   functions. */

/* the main program is called with no arguments */

main()
 {
  int IPeriod;                  /* variable that represents the period 
                                   under examination */

  InitStaticData();             /* get all input data from a file */

  InitMatrixOfStates();         /* initialize all states */

  InitBinUnitID();              /* initialize the binary representation 
                                   of the state values */
  InitOrderedUnitList();        /* order units by full load average cost */

/* calculate the LeastTotalCost for all periods */

  for (IPeriod = 1; IPeriod <= NumPeriods; IPeriod++)   
   {
    CalcLeastTotalCost(IPeriod); /* calculate the LeastTotalCost for all 
                                    states in this period */
   }
  TraceOptimalSchedule();       /* trace the optimal schedule */

  fclose(OutputFile);           /* close the output data file */
 }


/* This function reads the file ucdpin.dat to input constants into this 
   program */

/* input file format:

NumPeriods
LoadPattern[0] LoadPattern[1] ... LoadPattern[NumPeriods]

NumStates
NumCombos
NumLTLoad

NumUnits
Units[1].MinOutput Units[1].MaxOutput ... 
   Units[1].OnOffLine[PRES][1] Units[1].TimeOnOffLine[PRES][1] 
Units[2].MinOutput Units[2].MaxOutput ... 
   Units[2].OnOffLine[PRES][1] Units[2].TimeOnOffLine[PRES][1] 
.
.
.
Units[NumUnits].MinOutput Units[NumUnits].MaxOutput ... 
   Units[NumUnits].OnOffLine[PRES][1] Units[NumUnits].TimeOnOffLine[PRES][1] 

  */

void InitStaticData()
 {
  int   IPeriod;  /* index variable for period */
  int   IState;   /* index variable for state */
  int   IUnit;    /* index variable for unit */

  float MinOutput;
  float MaxOutput;
  float IncHeatRate;
  float NoLoadCost;
  float FullLoadAveCost;
  float FuelCost;
  int   MinUpTime;
  int   MinDownTime;
  float HotStartCost;
  float ColdStartCost;
  float ColdStartFixedCost;
  int   ColdStart;
  float ThermTimeConst;
  float MaintainTempCost;
  int   OnOffLine;
  int   TimeOnOffLine;

  FILE *InputFile;   /* pointer to the input file stream */

/* open the output data file (global variable) */

  OutputFile = fopen("ucdpout.dat", "w");

/* open the input data file */

  InputFile = fopen("ucdpin.dat", "r");

/* input the number of periods to be examined */

  fscanf(InputFile, "%d", &NumPeriods);

/* input the load pattern, one value for each period */

  for(IPeriod = 1; IPeriod <= NumPeriods; IPeriod++)
   {
    fscanf(InputFile, "%f", (LoadPattern + IPeriod));
   }

/* input the number of states to be examined */

  fscanf(InputFile, "%d", &NumStates);

/* input the number of combinations to be examined */

  fscanf(InputFile, "%d", &NumCombos);

/* input the number of generating units less than the number required 
   for load plus spinning reserve to be examined */

  fscanf(InputFile, "%d", &NumLTLoad);

/* input the number of generating units to be examined */

  fscanf(InputFile, "%d", &NumUnits);

/* input the generating unit descriptions */

  for(IUnit = 1; IUnit <= NumUnits; IUnit++)
   {
    fscanf(InputFile, "%f %f %f %f %f %f %d %d %f %f %f %d %f %f %d %d",
                         &MinOutput,
                         &MaxOutput,
                         &IncHeatRate,
                         &NoLoadCost,
                         &FullLoadAveCost,
                         &FuelCost,
                         &MinUpTime,
                         &MinDownTime,
                         &HotStartCost,
                         &ColdStartCost,
                         &ColdStartFixedCost,
                         &ColdStart,
                         &ThermTimeConst,
                         &MaintainTempCost,
                         &OnOffLine,
                         &TimeOnOffLine);

    Units[IUnit].MinOutput               = MinOutput;
    Units[IUnit].MaxOutput               = MaxOutput;
    Units[IUnit].IncHeatRate             = IncHeatRate;
    Units[IUnit].NoLoadCost              = NoLoadCost;
    Units[IUnit].FullLoadAveCost         = FullLoadAveCost;
    Units[IUnit].FuelCost                = FuelCost;
    Units[IUnit].MinUpTime               = MinUpTime;
    Units[IUnit].MinDownTime             = MinDownTime;
    Units[IUnit].HotStartCost            = HotStartCost;
    Units[IUnit].ColdStartCost           = ColdStartCost;
    Units[IUnit].ColdStartFixedCost      = ColdStartFixedCost;
    Units[IUnit].ColdStart               = ColdStart;
    Units[IUnit].ThermTimeConst          = ThermTimeConst;
    Units[IUnit].MaintainTempCost        = MaintainTempCost;
    Units[IUnit].OnOffLine[PRES][0]      = OnOffLine;
    Units[IUnit].TimeOnOffLine[PRES][0]  = TimeOnOffLine;

/* copy the on-off condition to the entire array for both the first 
   period and the initial period */

    Units[IUnit].OnOffLine[PREV][0]     = 
           Units[IUnit].OnOffLine[PRES][0];

    Units[IUnit].TimeOnOffLine[PREV][0] = 
           Units[IUnit].TimeOnOffLine[PRES][0];

    BinaryRep(0,0);

    for (IState = 1; IState < NumStates; IState++)
     {
      Units[IUnit].OnOffLine[PRES][IState]       =
             Units[IUnit].OnOffLine[PRES][0];

      Units[IUnit].OnOffLine[PREV][IState]       =
             Units[IUnit].OnOffLine[PRES][0];

      Units[IUnit].TimeOnOffLine[PRES][IState]   =
             Units[IUnit].TimeOnOffLine[PRES][0];

      Units[IUnit].TimeOnOffLine[PREV][IState]   =
             Units[IUnit].TimeOnOffLine[PRES][0];

      BinaryRep(0,IState);
     }
   }

  fclose(InputFile);
 }


/* The routine initializes the matrix of structures, so that the cost 
   to arrive at a state is set to INFINITY, a sufficiently large 
   floating point value, except in the 0th period.  Values are also 
   marked as "must run" or available for the 0th period */

void InitMatrixOfStates()
 {
  int IPeriod;    /* index variable for the period */
  int IState;     /* index variable for the state */
  int IUnit;      /* index variable for the generating unit */

/* initialize the values at each state in the matrix of states */

  for (IState = 0; IState < MAXSTATES; IState++)
   {
    for (IPeriod = 0; IPeriod <= MAXPERIODS; IPeriod++)
     {
      if(IPeriod == 0)
/* set the total cost to zero for the first stage */
       {
        States[IPeriod][IState].LeastTotalCost   = 0;
       }
      else
/* set the total cost to INFINITY */
       {
        States[IPeriod][IState].LeastTotalCost   = INFINITY;
        States[IPeriod][IState].GenStatus[0]     = '\0';
       }
/* set the backpointer to an impossible value */

      States[IPeriod][IState].BackPointer = -1;
     }
/* initialize the variables "must run" and "available" in the initial 
   period */

    for (IUnit = 1; IUnit <= NumUnits; IUnit++)
     {
/* mark unit as unavailable because of minimum down time */

      if ( (Units[IUnit].OnOffLine[PREV][IState] == OFFLINE) 
            &&
           (Units[IUnit].TimeOnOffLine[PREV][IState] <= 
            Units[IUnit].MinDownTime) )
       {
        Units[IUnit].Available[PREV][IState] = NO;
       }
       else
       {
        Units[IUnit].Available[PREV][IState] = YES;
       }
/* mark unit as must-run because of minimum up time */

      if ( (Units[IUnit].OnOffLine[PREV][IState] == ONLINE) 
            &&
           (Units[IUnit].TimeOnOffLine[PREV][IState] <= 
            Units[IUnit].MinUpTime))
       {
        Units[IUnit].MustRun[PREV][IState] = YES;
       }
       else
       {
        Units[IUnit].MustRun[PREV][IState] = NO;
       }
     }
   }
 }


/* This function initializes the matrix of unit identifiers, which 
   are just a binary representation of the value of IState. */

void InitBinUnitID()
 {
  int IState;     /* index variable for the state */
  int ICombo;     /* index variable for the combination */
  int TempState;  /* temporary variable for state values */
  int PowerOfTwo; /* temporary variable for the 2^(ICombo-1) */
  double DICombo;
  double DPowerOfTwo;

  for (IState = 0; IState < MAXSTATES; IState++)
   {
    TempState = IState;
    for (ICombo = NumCombos; ICombo > 0; ICombo--)
     {
      DICombo = ICombo - 1;
      DPowerOfTwo = pow(2.0,DICombo);
      PowerOfTwo  = DPowerOfTwo;
      if ( TempState >=  PowerOfTwo )
       {
        BinUnitID[IState][ICombo] = 1;
        TempState -= PowerOfTwo;
       }
       else
       {
        BinUnitID[IState][ICombo] = 0;
       }
     }
   }
 }


/* This function creates the list of generating units ordered by full load
   average cost */

void InitOrderedUnitList()
{
 int IUnit;  /* index variable for units */

 for (IUnit = 1; IUnit <= NumUnits; IUnit++)
  {
/* copy the full load average cost data and the pointer to the unit 
   with that value to the ordered units list */

   OrdUnits[IUnit].FullLoadAveCost = Units[IUnit].FullLoadAveCost;
   OrdUnits[IUnit].Pointer         = IUnit;
  }
  SortOrderedUnits(); /* sort the ordered units list */
}


/* This function is an application of the simple selection sort from 
   R. Sedgewick, "Algorithms", and sorts the array OrdUnits[] 
   by full load average cost */

void SortOrderedUnits() 
{
  int i,j;         /* index variables */
  int Min;         /* variable to hold the location of the minimum 
                      FullLoadAveCost value */
  float TempCost;    /* variable to hold the minimum value of FullLoadAveCost 
                      for each search cycle */
  int TempPointer; /* variable to hold the value of i corresponding to 
                      FullLoadAveCost for each search cycle */

  for (i = 1; i < NumUnits; i++)
   {
    Min = i;

    for (j = i+1; j <= NumUnits; j++)
     {
      if (OrdUnits[j].FullLoadAveCost < OrdUnits[Min].FullLoadAveCost)
       {
        Min = j;
       }
     }

    TempCost    = OrdUnits[Min].FullLoadAveCost;
    TempPointer = OrdUnits[Min].Pointer;

    OrdUnits[Min].FullLoadAveCost = OrdUnits[i].FullLoadAveCost;
    OrdUnits[Min].Pointer         = OrdUnits[i].Pointer;

    OrdUnits[i].FullLoadAveCost   = TempCost;
    OrdUnits[i].Pointer           = TempPointer;
   }

}
/* This function calculates the minimum cost to arrive at each state
   in a particular period. */

void CalcLeastTotalCost(IPeriod)
  int IPeriod;   /* period under examination */
 {
  int   IState;      /* index variable for the present state */
  int   PrevState;   /* index variable for the previous state */
  int   IUnit;       /* index variable for units */
  float Temp;        /* temporary storage value */
  float MaxCapacity; /* variable to calculate capacity */
  float ProdCost;    /* variable to store the calculated production cost */

  for (IState = 0; IState < NumStates; IState++)
   {
/* generate the ordered array of possible units */


   GeneratePossibleUnitList(IPeriod,IState);

/* determine whether or not this generating unit combination is feasible.
   it is not feasible if the maximum output is not greater than the 
   load pattern */

   MaxCapacity = 0.0;

   for (IUnit = 1; IUnit <= NumUnits; IUnit++)
    {
     if (Units[IUnit].OnOffLine[PRES][IState] == ONLINE)
      {
       MaxCapacity += Units[IUnit].MaxOutput;
      }
    }

   if (MaxCapacity >= LoadPattern[IPeriod])
    {
/* this generating combination is feasible so calculate the production 
   cost */

     ProdCost = ProductionCost(IPeriod,IState);

/* calculate a representation of the unit commitment at this period */


     BinaryRep(IPeriod,IState);


/* calculate the minimum value of least total cost */

     States[IPeriod][IState].LeastTotalCost = 
                               ProdCost +
                               TransitionCost(IPeriod,IState,0) +
                               States[IPeriod-1][0].LeastTotalCost;

     States[IPeriod][IState].BackPointer = 0;

     for (PrevState = 1; PrevState < NumStates; PrevState++)
      {
       Temp = ProdCost +
              TransitionCost(IPeriod,IState,PrevState) +
              States[IPeriod-1][PrevState].LeastTotalCost;

       if (Temp < States[IPeriod][IState].LeastTotalCost)
        {
         States[IPeriod][IState].LeastTotalCost = Temp;
         States[IPeriod][IState].BackPointer    = PrevState;
        }
      }
    }
    else  /* this generating unit combination is not feasible */
    {
     States[IPeriod][IState].LeastTotalCost = INFINITY;
    }
   }
/* move the present unit information to the previous array locations */

  for (IState = 0; IState < NumStates; IState++)
   {
    for (IUnit = 1; IUnit <= NumUnits; IUnit++)
     {
      Units[IUnit].OnOffLine[PREV][IState]     = 
             Units[IUnit].OnOffLine[PRES][IState];

      Units[IUnit].TimeOnOffLine[PREV][IState] = 
             Units[IUnit].TimeOnOffLine[PRES][IState];

      Units[IUnit].MustRun[PREV][IState]       = 
             Units[IUnit].MustRun[PRES][IState];

      Units[IUnit].Available[PREV][IState]     = 
             Units[IUnit].Available[PRES][IState];
     }
   }
 }


/* this function generates the ordered array of possible units */

void GeneratePossibleUnitList(IPeriod,IState)
  int IPeriod;        /* index variable for period */
  int IState;         /* index variable for state */
{
 int   IUnit;         /* index variable for units */
 int   ICombo;        /* index variable for combinations */
 int   IOrdUnits;     /* index variable to step through the ordered list */
 int   NumPossUnits;  /* number of possible units in a state */
 int   NumMustRun;    /* variable to keep track of the number of "must
                         run" generating units */
 float Capacity;      /* variable to calculate generating unit capacity */
 int   TempPossUnits[MAXNUMUNITS];  /* temporary array of possible 
                                      generating unit indexes */

 for (IUnit = 1; IUnit <= NumUnits; IUnit++)
  {
/* increment the TimeOnOffLine */

   Units[IUnit].TimeOnOffLine[PRES][IState]++;

/* mark unit as unavailable because of minimum down time */

   if ( (Units[IUnit].OnOffLine[PRES][IState] == OFFLINE) 
         &&
        (Units[IUnit].TimeOnOffLine[PRES][IState] <= 
         Units[IUnit].MinDownTime) )
    {
     Units[IUnit].Available[PRES][IState] = NO;
    }
    else
    {
     Units[IUnit].Available[PRES][IState] = YES;
    }
/* mark unit as must-run because of minimum up time */

   if ( (Units[IUnit].OnOffLine[PRES][IState] == ONLINE) 
         &&
        (Units[IUnit].TimeOnOffLine[PRES][IState] <= 
         Units[IUnit].MinUpTime))
    {
     Units[IUnit].MustRun[PRES][IState] = YES;
    }
    else
    {
     Units[IUnit].MustRun[PRES][IState] = NO;
    }
  }

/* generate the list of possible units */

  Capacity = 0.0;
  NumPossUnits = 0;

/* identify the "must run" units in this first pass through the Units 
   array */

  for (IUnit = 1; IUnit <= NumUnits; IUnit++)
   {
    if (Units[IUnit].MustRun[PRES][IState] == YES )
     {
      NumPossUnits++;
      TempPossUnits[NumPossUnits] = IUnit;
      Capacity += Units[IUnit].MaxOutput;
     }
   }

  NumMustRun = NumPossUnits;
  IOrdUnits = 0;

/* identify the number of units required to meet the load capacity
   plus spinning reserve requirement */

   while ( Capacity < LoadPattern[IPeriod] )
    {
     IOrdUnits++;

     if(IOrdUnits > NumUnits)
      {
       return;
      }

     if (Units[ OrdUnits[IOrdUnits].Pointer ].Available[PRES][IState] == YES
           &&
         Units[ OrdUnits[IOrdUnits].Pointer ].MustRun[PRES][IState]   == NO)
      {
       NumPossUnits++;
       TempPossUnits[NumPossUnits] = OrdUnits[IOrdUnits].Pointer;
       Capacity += Units[ OrdUnits[IOrdUnits].Pointer ].MaxOutput;
      }
    }

/* generate the first portion of the array of possible units.  make 
   sure all the "must run" generators are included */

   if((NumPossUnits - NumMustRun) < NumLTLoad)
    {
/* backtrack (NumPossUnits - NumMustRun) non "must run" and available 
   generators */

     for (IUnit = 1; IUnit <= (NumPossUnits - NumMustRun); IUnit++)
      {
       while (Units[ OrdUnits[IOrdUnits].Pointer ].Available[PRES][IState] 
                     == NO
                 ||
              Units[ OrdUnits[IOrdUnits].Pointer ].MustRun[PRES][IState] 
                     == YES)
        {
         IOrdUnits--; 
        }
       IOrdUnits--;
      }
     NumPossUnits = NumMustRun;
    }
    else
    {
     NumPossUnits -= NumLTLoad;

/* backtrack NumLTLoad non "must run" and available generators */

     for (IUnit = 1; IUnit <= NumLTLoad; IUnit++)
      {
       while (Units[ OrdUnits[IOrdUnits].Pointer ].Available[PRES][IState] 
                     == NO
                 ||
              Units[ OrdUnits[IOrdUnits].Pointer ].MustRun[PRES][IState] 
                     == YES)
        {
         IOrdUnits--; 
        }
       IOrdUnits--;
      }
    }
/* construct the possible unit list */

   for (IUnit = 1; IUnit <= NumPossUnits; IUnit++)
    {
     PossUnits[IUnit] = TempPossUnits[IUnit];

/* mark OnOffLine for units that have been off line to reflect that 
   they will be turned on */

     if (Units[ TempPossUnits[IUnit] ].OnOffLine[PRES][IState] == OFFLINE)
      {
       Units[ TempPossUnits[IUnit] ].OnOffLine[PRES][IState]     = ONLINE;
       Units[ TempPossUnits[IUnit] ].TimeOnOffLine[PRES][IState] = 0;
      }
    }

/* generate the combinations for the remainder of the array of possible
   units */

   for (ICombo = 1; ICombo <= NumCombos; ICombo++)
    {
/* advance to the next available unit not previously in "must run" */

     IOrdUnits++; 

     while (Units[ OrdUnits[IOrdUnits].Pointer ].Available[PRES][IState] 
                   == NO
               ||
            Units[ OrdUnits[IOrdUnits].Pointer ].MustRun[PRES][IState]   
                   == YES)
      {
       IOrdUnits++; 

       if(IOrdUnits > NumUnits)
        {
         return;
        }
      }

/* is this unit to be included in the combination? */

     if (BinUnitID[IState][ICombo] == YES)
      {
       NumPossUnits++; 

       PossUnits[NumPossUnits] = OrdUnits[IOrdUnits].Pointer;

/* mark OnOffLine for units that have been off line to reflect that 
   they will be turned on */

       if (Units[ PossUnits[NumPossUnits] ].OnOffLine[PRES][IState] == OFFLINE)
        {
         Units[ PossUnits[NumPossUnits] ].OnOffLine[PRES][IState]     = ONLINE;
         Units[ PossUnits[NumPossUnits] ].TimeOnOffLine[PRES][IState] = 0;
        }
      }
      else  /* this unit is not included in this combination */
      {
/* mark OnOffLine for units that have been on line to reflect that 
   they will be turned off */

       if (Units[ OrdUnits[IOrdUnits].Pointer ].OnOffLine[PRES][IState] 
                    == ONLINE)
        {
         Units[ OrdUnits[IOrdUnits].Pointer ].OnOffLine[PRES][IState]     
              = OFFLINE;
         Units[ OrdUnits[IOrdUnits].Pointer ].TimeOnOffLine[PRES][IState] 
              = 0;
        }
      }
    }
/* check all remaining units and mark OnOffLine for units that have been 
   on line to reflect that they will be turned off */

    for (IUnit = (IOrdUnits + 1); IUnit <= NumUnits; IUnit++);
     {
      if (Units[ OrdUnits[IOrdUnits].Pointer ].OnOffLine[PRES][IState] 
                   == ONLINE)
       {
        Units[ OrdUnits[IOrdUnits].Pointer ].OnOffLine[PRES][IState]     
              = OFFLINE;
        Units[ OrdUnits[IOrdUnits].Pointer ].TimeOnOffLine[PRES][IState] 
              = 0;
       }
     }
}


/* This function calculates a string representation of the unit commitment 
   at a state within a period */

void BinaryRep(IPeriod,IState)
  int IPeriod;   /* present period under examination */
  int IState;    /* present state under examination */
{
  int IUnit;     /* index variable for generating units */

  for (IUnit = 1; IUnit <= NumUnits; IUnit++)
   {
    if(Units[IUnit].OnOffLine[PRES][IState] == ONLINE)
     {
      States[IPeriod][IState].GenStatus[IUnit] = '1';
     }
     else
     {
      States[IPeriod][IState].GenStatus[IUnit] = '0';
     }
   }
/* add an end-of-string marker so this string can be printed */

  States[IPeriod][IState].GenStatus[(NumUnits+1)] = '\0';
}


/* This function calculates and returns the value of 
   ProductionCost(IPeriod,IState,) after an economic dispatch is 
   calculated to determine the optimal production */

float ProductionCost(IPeriod,IState)
  int IPeriod;    /* present period being examined */
  int IState;     /* present state being examined */
{
  int   IUnit;    /* index variables for units */
  float ProdCost; /* variable to calculate sum of production costs */

/* calculate an economic dispatch for this state and period */

  EconomicDispatch(IPeriod,IState);

/* calculate the production cost */

  ProdCost = 0.0;

  for (IUnit = 1; IUnit <= NumUnits; IUnit++)
   {
    if (Units[IUnit].OnOffLine[PRES][IState] == ONLINE)
     {
      ProdCost += (( (Units[IUnit].NoLoadCost + 
                      (Units[IUnit].UnitProduction[IState] * 
                       Units[IUnit].IncHeatRate)) ) *
                   Units[IUnit].FuelCost);
     }
   }
  return ProdCost;
}


/* This function calculates an economic dispatch for linear input-output
   production cost curves for generating units */

void EconomicDispatch(IPeriod,IState)
  int IPeriod;       /* present period being examined */
  int IState;        /* present state being examined */
{
  int   IUnit;       /* index variable for unit */
  int   IOrdUnits;   /* index variable to step through the ordered list */
  float MaxCapacity; /* variable to calculate capacity */

/* set unit production to max output for all units and calculate the 
   maximum capacity */

   MaxCapacity = 0.0;

   for (IUnit = 1; IUnit <= NumUnits; IUnit++)
    {
     if (Units[IUnit].OnOffLine[PRES][IState] == ONLINE)
      {
       Units[IUnit].UnitProduction[IState] = Units[IUnit].MaxOutput;
       MaxCapacity += Units[IUnit].MaxOutput;
      }
    }

/* reduce the unit production for those units with the highest full 
   load cost until the total capacity equals the load pattern */

   IOrdUnits = NumUnits + 1;

   while (MaxCapacity > LoadPattern[IPeriod])
    {
/* locate the next online generator */

     IOrdUnits--;

     while (Units [ OrdUnits[IOrdUnits].Pointer ].OnOffLine[PRES][IState]
              == OFFLINE)
      {
       IOrdUnits--;
      }

/* adjust this unit's output to reduce the difference in maximum 
   capacity and load pattern - the unit's capacity cannot be reduced 
   below its minimum output */

     if ( (MaxCapacity - LoadPattern[IPeriod])
           >
         (Units [ OrdUnits[IOrdUnits].Pointer ].MaxOutput -
          Units [ OrdUnits[IOrdUnits].Pointer ].MinOutput) )

      {
/* this unit cannot absorb all the difference in capacity - so only 
   reduce the unit production to minimum output */

       Units[ OrdUnits[IOrdUnits].Pointer ].UnitProduction[IState] 
              = Units[ OrdUnits[IOrdUnits].Pointer ].MinOutput;

       MaxCapacity -=
            (Units [ OrdUnits[IOrdUnits].Pointer ].MaxOutput -
             Units [ OrdUnits[IOrdUnits].Pointer ].MinOutput);
      }
      else
      {
/* this unit can absorb all the remaining difference in capacity */

       Units[ OrdUnits[IOrdUnits].Pointer ].UnitProduction[IState] 
             -= (MaxCapacity - LoadPattern[IPeriod]);

       MaxCapacity = LoadPattern[IPeriod];
      }

    }
}


/* This function calculates and returns the value of 
   TransitionCost(IPeriod,IState,PrevState) */

float TransitionCost(IPeriod,IState,PrevState) 
  int IPeriod;    /* present period being examined */
  int IState;     /* present state being examined */
  int PrevState;  /* state within previous period being examined */
{
  int   IUnit;       /* index variable for units */
  int   Feasible;    /* indicator variable to determine whether or not 
                        this transition combination is feasible */
  float TransCost;   /* variable to calculate sum of transition costs */
  float CoolingCost; /* variable to calculate cooling cost */
  float BankingCost; /* variable to calculate banking cost */

/* determine if this transition is feasible */

  Feasible = YES;

  for (IUnit = 1; IUnit <= NumUnits; IUnit++)
   {
/* determine if a unit that is "must run" in the previous period should 
   also be "must run" in this period */

    if ( ((Units[IUnit].MustRun[PREV][PrevState] == YES)  &&
          ((Units[IUnit].TimeOnOffLine[PREV][PrevState] + 1) <=
           Units[IUnit].MinUpTime))

         &&

         (Units[IUnit].MustRun[PRES][IState] == NO) )
     {
      Feasible = NO;
     }
/* determine if a unit that is unavailable in the previous period 
   should also be unavailable in this period */

    if ( ((Units[IUnit].Available[PREV][PrevState] == NO)  &&
          ((Units[IUnit].TimeOnOffLine[PREV][PrevState] + 1) <=
           Units[IUnit].MinDownTime))

         &&

         (Units[IUnit].Available[PRES][IState] == YES) )
     {
      Feasible = NO;
     }
   }
/* if this transition is feasible calculate the transition cost 
   otherwise return INFINITY */

  if (Feasible == YES)
   {
/* calculate the transition cost - shutdown costs will be ignored */

    TransCost = 0.0;

    for (IUnit = 1; IUnit <= NumUnits; IUnit++)
     {
/* calculate startup costs - unit will be online for only 0 hours */

      if ( (Units[IUnit].OnOffLine[PRES][IState]    == ONLINE)
            &&
           (Units[IUnit].OnOffLine[PREV][PrevState] == OFFLINE) )
       {
/* determine whether or not this will be a cold start or hot start */
        if ( (Units[IUnit].TimeOnOffLine[PREV][PrevState] + 1) >= 
              Units[IUnit].ColdStart)
         {
          TransCost += Units[IUnit].ColdStartFixedCost;
         }
         else
         {
           CoolingCost = (Units[IUnit].ColdStartCost *
                          (1.0 - exp( -1.0 * 
                           ((Units[IUnit].TimeOnOffLine[PREV][PrevState] + 1) /
                             Units[IUnit].ThermTimeConst))) *
                          Units[IUnit].FuelCost) +
                         Units[IUnit].ColdStartFixedCost;

           BankingCost = (Units[IUnit].MaintainTempCost *
                          (Units[IUnit].TimeOnOffLine[PREV][PrevState] + 1) *
                          Units[IUnit].IncHeatRate) +
                         Units[IUnit].ColdStartFixedCost;

/* transition cost for this unit will be the minimum of cooling and
   banking costs */

           if(CoolingCost < BankingCost)
            {
             TransCost += CoolingCost;
            }
            else
            {
             TransCost += BankingCost;
            }
         }
       }
     }
   }
   else /* this transition was not feasible */
   {
    TransCost = INFINITY;
   }

  return TransCost;
}


/* This function traces the optimal schedule by starting at the least
   cost final state, and works backward until the 0th state is reached. */

void TraceOptimalSchedule()
 {
  int  IPeriod;         /* index variable for period */
  int  IState;          /* index variable for state */
  int  IUnit;           /* index variable for units */
  int  TempBackPointer; /* temporary variable for backpointer */
  float TempCost;        /* value used to find the minimum cost final state */
  int  OptimalSchedule[MAXPERIODS+1]; /* array of states in the optimal 
                                        schedule */

/* find the minimum cost state in the last period */

  TempBackPointer = 0;
  TempCost = States[NumPeriods][0].LeastTotalCost;

  for (IState = 1; IState < NumStates; IState++)
   {
    if (States[NumPeriods][IState].LeastTotalCost < TempCost)
     {
      TempBackPointer = IState;
      TempCost = States[NumPeriods][IState].LeastTotalCost;
     }
   }
/* record the location of the minimum cost state in the last period */

  OptimalSchedule[NumPeriods] = TempBackPointer;

/* trace through the remaining periods */

  for (IPeriod = (NumPeriods-1); IPeriod >= 0 ; IPeriod--)
   {
    OptimalSchedule[IPeriod] = 
     States[IPeriod+1][ OptimalSchedule[IPeriod+1] ].BackPointer;
   }

/* print information on the trace */
    fprintf(OutputFile,"OPTIMAL SCHEDULE\n\n");

    fprintf(OutputFile,"              ");
    for (IUnit = 1; IUnit <= NumUnits; IUnit++)
     {
      fprintf(OutputFile,"%1d",(IUnit % 10));
     }
    fprintf(OutputFile,"\n");

    for (IPeriod = 0; IPeriod <= NumPeriods; IPeriod++)
     {
      fprintf(OutputFile,"period #%3d - ",IPeriod);
      fflush(OutputFile);
      for (IUnit = 1; IUnit <= NumUnits; IUnit++)
       {
         fprintf(OutputFile,"%c",
                 States[IPeriod][ OptimalSchedule[IPeriod] ].GenStatus[IUnit]);
         fflush(OutputFile);
       }
      fprintf(OutputFile," - %10.2f \n",
              States[IPeriod][ OptimalSchedule[IPeriod] ].LeastTotalCost);
      fflush(OutputFile);
     }
 }


